#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <sys/select.h>
#include <unistd.h>
#include <cairo/cairo.h>
#include <cairo/cairo-xlib.h>
#include <string.h>
#include <math.h>

#include "solve.h"
#include "solve2.h"
#include "solve3.h"
#include "last-ring.h"
#include "endgame.h"

move_t *moves;
int moves_size;

void move_dump(void)
{
  int i;
  for (i = 0; i < moves_size; i++)
    printf("%d %d %d\n", moves[i].direction, moves[i].x, moves[i].y);

  FILE *f = fopen("/tmp/moves.txt", "w"); if (f == NULL) exit(1);
  for (i = 0; i < moves_size; i++)
    fprintf(f, "%d %d %d\n", moves[i].direction, moves[i].x, moves[i].y);
  fclose(f);

  system("./translate-move < /tmp/moves.txt 2>/dev/null");
}

void move_push(int direction, int x, int y)
{
  moves_size++;
  move_t *new_moves = realloc(moves, moves_size * sizeof(move_t));
  if (new_moves == NULL) { move_dump(); exit(1); }
  moves = new_moves;
  moves[moves_size - 1].direction = direction;
  moves[moves_size - 1].x = x;
  moves[moves_size - 1].y = y;
}

/* end of moves store / dump */

#define WIDTH (256*2*2)
#define HEIGHT (256*2*2)

unsigned int grid[2*N-1][2*N-1];

point_t cursor = { -1000, -1000 };

void set_grid(point_t p, int v)
{
  grid[(int)p.y][(int)p.x] = v;
}

int get_grid(point_t p)
{
  return grid[(int)p.y][(int)p.x];
}

enum {
  LEFT, RIGHT
} rotation = RIGHT;

point_t grid_pos(int line, int column)
{
  point_t ret;
  int start_x = 10 + abs(line - N + 1) * 5;

  ret.y = 10 + line * 10 * cos(M_PI / 6);
  ret.x = 10 * column + start_x;

  return ret;
}

int columns_per_line(int line)
{
  return N + N - 1 - abs(line - N + 1);
}

int is_grid(int line, int column)
{
  return line >= 0 && line < N * 2 - 1 && column >= 0 && column < columns_per_line(line);
}

void init_grid(void)
{
  int init[] = { INIT };
  int p = 0;
  for (int line = -1; line < 2 * N; line++)
    for (int column = -1; column < 2 * N; column++) {
      if (!is_grid(line, column)) continue;
      grid[line][column] = init[p];
      p++;
    }
printf("final p %d\n", p);
}

/* this works only for border positions, not arbitrary over all the 2D plane */
point_t canonical_position(point_t p)
{
  if (is_grid(p.y, p.x)) return p;

  /* upper horizontal line */
  if (p.y == -1 && p.x >= 0)
    return (point_t) { .x = p.x, .y = (N - 1) * 2 };

  /* upper left line */
  if (p.x == -1 && p.y < N - 1)
    return (point_t) { .x = columns_per_line(p.y + N) - 1, .y = p.y + N };

  /* lower left line */
  if (p.x == -1 && p.y <= (N - 1) * 2)
    return (point_t) { .x = columns_per_line(p.y - (N - 1)) - 1, .y = p.y - (N - 1) };

  /* lower horizontal line */
  if (p.y == (N - 1) * 2 + 1 && p.x < N - 1)
    return (point_t) { .x = p.x + 1, .y = 0 };

  /* upper right line */
  if (p.y <= N - 1)
    return (point_t) { .x = 0, .y = p.y + N - 1 };

  /* lower right line */
  return (point_t) { .x = 0, .y = p.y - N };
}

void positions_around(point_t *around, point_t p)
{
  int ax, ay;
  int bx, by;
  int cx, cy;
  int a, b, c;

  if ((int)p.y % 2) {
    /* point to bottom
     * \a/
     * c|b
     */
    /* get coordinates (x, y) of a */
    ay = ((int)p.y - 1) / 2 - 1;
    ax = p.x;
    if (ay < N - 1)
      ax--;
    point_t p = canonical_position((point_t) { ax, ay });
    around[0] = p;
    a = grid[(int)p.y][(int)p.x];
    /* get b */
    by = ay + 1;
    bx = ax;
    if (ay < N - 1)
      bx++;
    p = canonical_position((point_t) { bx, by });
    around[1] = p;
    b = grid[(int)p.y][(int)p.x];
    /* get c */
    cx = bx - 1;
    cy = by;
    p = canonical_position((point_t) { cx, cy });
    around[2] = p;
    c = grid[(int)p.y][(int)p.x];
  } else {
    /* points to top
     * b|c
     * /a\
     */
    /* get a */
    ay = (int)p.y / 2;
    ax = p.x;
    if (ay > N - 1)
      ax--;
    point_t p = canonical_position((point_t) { ax, ay });
    around[0] = p;
    a = grid[(int)p.y][(int)p.x];
    /* get b */
    by = ay - 1;
    bx = ax;
    if (ay < N)
      bx--;
    p = canonical_position((point_t) { bx, by });
    around[1] = p;
    b = grid[(int)p.y][(int)p.x];
    /* get c */
    cx = bx + 1;
    cy = by;
    p = canonical_position((point_t) { cx, cy });
    around[2] = p;
    c = grid[(int)p.y][(int)p.x];
  }
(void)a; (void)b; (void)c;
#if 0
printf("p %d %d a %d (x/y %d %d) b %d (x/y %d %d) c %d (x/y %d %d)\n", (int)p.x, (int)p.y, a, ax, ay, b, bx, by, c, cx, cy);
#endif
}

int valid_cursor(point_t p, unsigned int grid[2*N-1][2*N-1])
{
  if (p.y < 0 || p.y > N * 4 - 1
      || p.x < 0)
    return 0;

  /* this is not the same as columns_per_line(), this is relative to the cursor, not the board */
  int columns_per_line = N + N - abs(((int)p.y + 1) / 2 - N);
  if (!(p.x < columns_per_line))
    return 0;

  /* one of the position around cursor must be 0 */
  point_t around[3];
  positions_around(around, p);
  return    grid[(int)around[0].y][(int)around[0].x] == 0
         || grid[(int)around[1].y][(int)around[1].x] == 0
         || grid[(int)around[2].y][(int)around[2].x] == 0;
}

void set_cursor(point_t c)
{
  cursor = c;
}

point_t cursor_to_screen(point_t cursor)
{
  float line;
  float column;
  float start_column;

  start_column = 10 + abs((int)cursor.y / 2 - N + 1) * 5;
  start_column -= 5 * ((int)cursor.y % 2);
  if ((int)cursor.y / 2 >= N && (int)cursor.y % 2 == 0)
    start_column -= 10;

  line = 5 + ((int)cursor.y / 2) * 10 * cos(M_PI / 6) + ((int)cursor.y % 2) * 2;
  column = cursor.x * 10 + start_column;

  return (point_t){ column * 4, line * 4};
}

int delta_x = 0;
int delta_y = 0;

point_t screen_to_cursor(point_t pixel)
{
  pixel.x -= delta_x;
  pixel.y -= delta_y;

  pixel.x -= 30;
  pixel.y -= 30;

  int x = pixel.x / 4;
  int y = pixel.y / 4;

  /* get cursor y, reverse cursor_to_screen() */
  int cursor_y = (y - 5) / (5 * cos(M_PI / 6));
  /* deal with % 2 by taking nearest of (cursor_y, cursor_y + 1) */
  point_t a = cursor_to_screen((point_t) { 0, cursor_y });
  point_t b = cursor_to_screen((point_t) { 0, cursor_y + 1 });
  if (abs(b.y - pixel.y) < abs(a.y - pixel.y))
    cursor_y++;

  /* get cursor x */
  int start_column = 10 + abs(cursor_y / 2 - N + 1) * 5;
  start_column -= 5 * (cursor_y % 2);
  if (cursor_y / 2 >= N && cursor_y % 2 == 0)
    start_column -= 10;
  int cursor_x = (x - start_column) / 10;
  a = cursor_to_screen((point_t) { cursor_x, cursor_y });
  b = cursor_to_screen((point_t) { cursor_x + 1, cursor_y });
  if (abs(b.x - pixel.x) < abs(a.x - pixel.x))
    cursor_x++;

  /* maybe not the nearest, let's also try the 8 points around */
  int nearest_x = cursor_x;
  int nearest_y = cursor_y;
  point_t nearest = cursor_to_screen((point_t) { nearest_x, nearest_y });
  int nearest_dist = (pixel.x - nearest.x) * (pixel.x - nearest.x)
                   + (pixel.y - nearest.y) * (pixel.y - nearest.y);
  for (int x = cursor_x - 1; x <= cursor_x + 1; x++)
    for (int y = cursor_y - 1; y <= cursor_y + 1; y++) {
      point_t current = cursor_to_screen((point_t) { x, y });
      int current_dist = (pixel.x - current.x) * (pixel.x - current.x)
                       + (pixel.y - current.y) * (pixel.y - current.y);
      if (current_dist < nearest_dist) {
        nearest_x = x;
        nearest_y = y;
        nearest_dist = current_dist;
      }
    }

//printf("%d %d\n", nearest_x, nearest_y);

  /* out of grid? no cursor */
  if (!valid_cursor((point_t) { nearest_x, nearest_y }, grid)) {
    nearest_x = -1000;
    nearest_y = -1000;
  }

  return (point_t){ nearest_x, nearest_y };
}

cairo_surface_t *cs;
cairo_surface_t *csout;

void expose(Display *d, Window w, Pixmap p)
{
  XCopyArea(d, p, w, DefaultGC(d, DefaultScreen(d)), 0,0,
            WIDTH, HEIGHT, 0,0);
}

void paint_number(cairo_t *c, int v, int center_x, int center_y)
{
  if (v == 0) return;
  char s[32];
  sprintf(s, "%d", v);
  cairo_text_extents_t extents;
  cairo_text_extents(c, s, &extents);
  /* typedef struct {
       double x_bearing;
       double y_bearing;
       double width;
       double height;
       double x_advance;
       double y_advance;
     } cairo_text_extents_t;
   */
  cairo_move_to(c, center_x - extents.width / 2 - extents.x_bearing, center_y + extents.height / 2);
  cairo_show_text(c, s);
}

int repeated_number(int line, int column)
{
  point_t p = canonical_position((point_t){ .x = column, .y = line });
  return grid[(int)p.y][(int)p.x];
}

void paint(void)
{
  cairo_t *c;

  c=cairo_create(cs);

  cairo_set_font_size(c, 18);

  /* clear screen */
  cairo_rectangle(c, 0, 0, WIDTH, HEIGHT);
  cairo_set_source_rgb(c, 0.1, 0.1, 0.3);
  cairo_fill(c);

  cairo_translate(c, delta_x, delta_y);
  cairo_translate(c, 30, 30);

  /* draw repeated grid parts around main grid */
  cairo_set_source_rgb(c, .4, .4, .8);
  for (int line = -1; line < 2 * N; line++) {
    int inc = line == -1 || line == 2 * N - 1 ? 1 : columns_per_line(line) + 1;
    for (int column = -1; column < columns_per_line(line) + 1; column += inc) {
      point_t p = grid_pos(line, column);
      p.x *= 4;
      p.y *= 4;
      //cairo_arc(c, p.x, p.y, 3, 0, 2 * M_PI);
      paint_number(c, repeated_number(line, column), p.x, p.y);
      //cairo_fill(c);
      float x1, y1;
      for (int i = 0; i < 6; i++) {
        x1 = cos(2 * M_PI * i / 6 + M_PI / 2);
        y1 = sin(2 * M_PI * i / 6 + M_PI / 2);
        float v = 1 / cos(M_PI / 6);
        x1 *=2 * 10 * v; y1 *= 2 * 10 * v;
        x1 += p.x; y1 += p.y;
        if (i == 0) cairo_move_to(c, x1, y1); else cairo_line_to(c, x1, y1);
      }
      cairo_close_path(c);
      cairo_stroke(c);
    }
  }

  /* draw grid */
  cairo_set_source_rgb(c, .6, .6, .9);
  for (int line = 0; line < 2 * N - 1; line++)
    for (int column = 0; column < 2 * N - 1; column++) {
      if (!is_grid(line, column)) continue;
      point_t p = grid_pos(line, column);
      p.x *= 4;
      p.y *= 4;
      //cairo_arc(c, p.x, p.y, 3, 0, 2 * M_PI);
      paint_number(c, grid[line][column], p.x, p.y);
      //cairo_fill(c);
      float x1, y1;
      for (int i = 0; i < 6; i++) {
        x1 = cos(2 * M_PI * i / 6 + M_PI / 2);
        y1 = sin(2 * M_PI * i / 6 + M_PI / 2);
        float v = 1 / cos(M_PI / 6);
        x1 *=2 * 10 * v; y1 *= 2 * 10 * v;
        x1 += p.x; y1 += p.y;
        if (i == 0) cairo_move_to(c, x1, y1); else cairo_line_to(c, x1, y1);
      }
      cairo_close_path(c);
      cairo_stroke(c);
    }

  /* draw cursor */
  /* ugly, todo: cleanup this code */
  point_t cursor_screen = cursor_to_screen(cursor);
  if (cursor.x == -1000) goto no_cursor;
  cairo_save(c);
  cairo_set_line_width(c, 3);
  if (rotation == LEFT) {
    cairo_set_source_rgb(c, .4, .6, .4);
    /* cursor circle (not full) */
    cairo_arc(c, cursor_screen.x, cursor_screen.y, 10,
              M_PI / 2 + M_PI / 7, M_PI / 2 - M_PI / 7);
    /* cursor arrow */
    point_t p1, p2, p3;
    /* p2 is at end of circle */
    double p2x, p2y;
    cairo_get_current_point(c, &p2x, &p2y);
    p2.x = p2x;
    p2.y = p2y;
    /* p1 away from cicle (angle choseen to "look good") */
    float angle = M_PI / 2 - M_PI / 7
                - M_PI / 3;
    float cc = cos(angle);
    float ss = sin(angle);
    p1 = (point_t) { p2.x + 6 * cc,
                     p2.y + 6 * ss };
    /* p3 inside circle (angle chosen to "look good") */
    angle = M_PI / 2 - M_PI / 7
          + M_PI
          + M_PI / 6;
    cc = cos(angle);
    ss = sin(angle);
    p3 = (point_t) { p2.x + 6 * cc,
                     p2.y + 6 * ss };
    cairo_move_to(c, p1.x, p1.y);
    cairo_line_to(c, p2.x, p2.y);
    cairo_line_to(c, p3.x, p3.y);
  } else {
    cairo_set_source_rgb(c, .8, .5, .5);
    /* cursor circle (not full) */
    cairo_arc_negative(c, cursor_screen.x, cursor_screen.y, 10,
              M_PI / 2 - M_PI / 7 , M_PI / 2 + M_PI / 7);
    /* cursor arrow */
    point_t p1, p2, p3;
    /* p2 is at end of circle */
    double p2x, p2y;
    cairo_get_current_point(c, &p2x, &p2y);
    p2.x = p2x;
    p2.y = p2y;
    /* p1 away from cicle (angle choseen to "look good") */
    float angle = M_PI / 2 + M_PI / 7
                + M_PI / 3;
    float cc = cos(angle);
    float ss = sin(angle);
    p1 = (point_t) { p2.x + 6 * cc,
                     p2.y + 6 * ss };
    /* p3 inside circle (angle chosen to "look good") */
    angle = M_PI / 2 + M_PI / 7
          - M_PI
          - M_PI / 6;
    cc = cos(angle);
    ss = sin(angle);
    p3 = (point_t) { p2.x + 6 * cc,
                     p2.y + 6 * ss };
    cairo_move_to(c, p1.x, p1.y);
    cairo_line_to(c, p2.x, p2.y);
    cairo_line_to(c, p3.x, p3.y);
  }
  cairo_stroke(c);
  cairo_restore(c);
no_cursor:

  cairo_destroy(c);

  /* and put it all to the xpixmap */
  c=cairo_create(csout);
  cairo_set_source_surface(c, cs, 0, 0);
  cairo_rectangle(c, 0, 0, WIDTH, HEIGHT);
  cairo_paint(c);
  cairo_destroy(c);
}

void usage(void)
{
  printf("options:\n");
  exit(0);
}

/* replay logic */
FILE *replay;
void start_replay(void)
{
  if (replay != NULL) fclose(replay);
  replay = fopen("/tmp/moves.txt", "r");
  if (replay == NULL) printf("could not open replay file /tmp/moves.txt\n");
  printf("starting replay of /tmp/moves.txt, press 'b' to read moves\n");
}

move_t replay_next(void)
{
  if (replay == NULL) { printf("replay file not opened or finished\n"); return (move_t){.x=-1}; }
  move_t ret;
  if (fscanf(replay, "%d %d %d", &ret.direction, &ret.x, &ret.y) != 3) {
    fclose(replay);
    replay = NULL;
    return (move_t) { .x = -1 };
  }
  printf("replay x %d y %d direction %d\n", ret.x, ret.y, ret.direction);
  return ret;
}

int main(int n, char **v)
{
  Display *d;
  Window w;
  Pixmap p;
  fd_set r;
  int redraw;
  int repaint;
  int i;
  point_t pointer_position = (point_t) { -1000, -1000 };
  moves_t m;

  for (i = 1; i < n; i++) {
  }

  init_grid();

  cs = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,
                                  WIDTH, HEIGHT);

  d = XOpenDisplay(0); if (!d) abort();
  w = XCreateSimpleWindow(d, DefaultRootWindow(d), 0, 0,
                          WIDTH, HEIGHT, 0, 0, 0);
  XSelectInput(d, w, KeyPressMask | ExposureMask | PointerMotionMask
                     | ButtonPressMask | EnterWindowMask | LeaveWindowMask);
  XMapWindow(d, w);

  p = XCreatePixmap(d, w, WIDTH,HEIGHT,
                    DefaultDepth(d, DefaultScreen(d)));
  csout = cairo_xlib_surface_create(d, p, DefaultVisual(d, DefaultScreen(d)),
                                    WIDTH, HEIGHT);

  /* get a smaller cursor */
  Cursor c = XCreateFontCursor(d, XC_plus);
  XDefineCursor(d, w, c);
  XFreeCursor(d, c);

  paint();
  redraw = 0;
  repaint = 0;

  while (1) {
    XEvent ev;
    while (XPending(d)) {
      XNextEvent(d, &ev);
      switch (ev.type) {
      case Expose: redraw = 1; break;
      case KeyPress: {
        char buf[65];
        KeySym ks;
        XLookupString(&ev.xkey, buf, 64, &ks, NULL);
        switch (ks) {
          case ' ': rotation = 1 - rotation; repaint = 1; break;
          case 'a': start_replay(); break;
          case 'b': {
            move_t mv = replay_next();
            if (mv.x != -1) {
              m.size = 1;
              m.moves = malloc(sizeof(move_t)); if (m.moves == NULL) abort();
              m.moves[0] = mv;
              goto do_moves;
            }
            break;
          }
          case 'c': {
            /* find hole */
            int x, y;
            for (y = 0; y < N*2-1; y++) for (x = 0; x < 2*N-1; x++)
              if (is_grid(y, x) && grid[y][x] == 0) goto found_hole;
            found_hole:
            int center_x = WIDTH / 2;
            int center_y = HEIGHT / 2;
            point_t p = grid_pos(y, x);
            p.x *= 4;
            p.y *= 4;
            int hole_x = p.x;
            int hole_y = p.y;
            delta_x = center_x - hole_x;
            delta_y = center_y - hole_y;
            set_cursor(screen_to_cursor(pointer_position));
            repaint = 1;
            break;
          }
          case 'r':
            delta_x = delta_y = 0;
            set_cursor(screen_to_cursor(pointer_position));
            repaint = 1;
            break;
          case 'd': move_dump(); break;
          case 's': {
            /* first call of solve(): rotate as you want, next calls: respect rotation */
            m = solve(moves_size == 0 ? -1 : rotation);
do_moves:
            if (m.size > 0) {
              rotation = m.moves[0].direction;
              for (int i = 0; i < m.size; i++) {
                point_t p[3];
                positions_around(p, (point_t){.x = m.moves[i].x, .y = m.moves[i].y});
                move_push(m.moves[i].direction, m.moves[i].x, m.moves[i].y);
                if (m.moves[i].direction == LEFT) {
                  int v = get_grid(p[0]);
                  set_grid(p[0], get_grid(p[2]));
                  set_grid(p[2], get_grid(p[1]));
                  set_grid(p[1], v);
                } else {
                  int v = get_grid(p[0]);
                  set_grid(p[0], get_grid(p[1]));
                  set_grid(p[1], get_grid(p[2]));
                  set_grid(p[2], v);
                }
              }
              rotation = 1 - m.moves[m.size - 1].direction;
              set_cursor(screen_to_cursor(pointer_position));
              repaint = 1;
            }
            free(m.moves);
            break;
          }
          case 't':
            if (moves_size == 0) { printf("error: can't do t\n"); break; }
            m = get_last_moves(grid, rotation);
            goto do_moves;
          case 'z':
            m = endgame(grid, rotation);
            goto do_moves;
          case 'u':
            /* first call of solve2(): rotate as you want, next calls: respect rotation */
            //moves_t m = solve2(moves_size == 0 ? -1 : rotation);
            m = solve2(rotation);
            goto do_moves;
          case 'v':
            m = solve3(rotation);
            goto do_moves;
          case XK_Return: goto rotate;
          //case XK_Right: case 'n': printf("n"); fflush(stdout); break;
          //case XK_Left:  case 'p': printf("p"); fflush(stdout); break;
          case XK_Escape: case 'q': move_dump(); goto end;
        }
        break;
      }
      case ButtonPress: {
        if (ev.xbutton.button != Button1) break;
        point_t old_cursor = cursor;
        pointer_position = (point_t){ ev.xmotion.x, ev.xmotion.y };
        set_cursor(screen_to_cursor(pointer_position));
        repaint |= old_cursor.x != cursor.x || old_cursor.y != cursor.y;
rotate:
        /* rotate is cursor valid */
        if (valid_cursor(cursor, grid)) {
printf("%d %d\n", (int)cursor.x, (int)cursor.y);
          point_t p[3];
          positions_around(p, cursor);
          move_push(rotation, cursor.x, cursor.y);
          if (rotation == LEFT) {
            int v = get_grid(p[0]);
            set_grid(p[0], get_grid(p[2]));
            set_grid(p[2], get_grid(p[1]));
            set_grid(p[1], v);
          } else {
            int v = get_grid(p[0]);
            set_grid(p[0], get_grid(p[1]));
            set_grid(p[1], get_grid(p[2]));
            set_grid(p[2], v);
          }
          rotation = 1 - rotation;
          repaint = 1;
        }
        break;
      }
      case MotionNotify: {
        //printf("move to %d %d\n", ev.xmotion.x, ev.xmotion.y);
        point_t old_cursor = cursor;
        pointer_position = (point_t){ ev.xmotion.x, ev.xmotion.y };
        set_cursor(screen_to_cursor(pointer_position));
        repaint |= old_cursor.x != cursor.x || old_cursor.y != cursor.y;
        break;
      }
      case EnterNotify: {
        //printf("enter at %d %d\n", ev.xcrossing.x, ev.xcrossing.y);
        point_t old_cursor = cursor;
        pointer_position = (point_t) { ev.xcrossing.x, ev.xcrossing.y };
        set_cursor(screen_to_cursor(pointer_position));
        repaint |= old_cursor.x != cursor.x || old_cursor.y != cursor.y;
        break;
      }
      case LeaveNotify: {
        //printf("leave at %d %d\n", ev.xcrossing.x, ev.xcrossing.y);
        point_t old_cursor = cursor;
        pointer_position = (point_t) { -1000, -1000 };
        set_cursor(screen_to_cursor(pointer_position));
        repaint |= old_cursor.x != cursor.x || old_cursor.y != cursor.y;
        break;
      }
      default: break;
      }
    }
    if (repaint) { paint(); repaint = 0; redraw = 1; }
    if (redraw) { expose(d, w, p); redraw = 0; }
    FD_ZERO(&r);
    FD_SET(ConnectionNumber(d), &r);
    XFlush(d);
    if (select(ConnectionNumber(d)+1, &r, 0, 0, 0) == -1) abort();
  }

end:
  XFreePixmap(d, p);
  XCloseDisplay(d);
  cairo_surface_destroy(cs);
  cairo_surface_destroy(csout);
  cairo_debug_reset_static_data();

  return 0;
}
