#include <stdio.h>
#include <stdlib.h>

/* last ring is (this is also the target state):
      /\/\
     |0|1|
     /\/\/\
    |2|3|4|
     \/\/\/
     |5|6|
      \/\/
  we have 7*5*4*3*2 possible positions = 5040

  12 possible moves (corners of 3), numbered as:

       1
     6/\2
     | |
     5\/3
      4

  positive if rotate in clock direction, negative in counterclock
 */

/* stones to move, for each move, in clock direction
 * (hopefully correct, if not, then counterclock)
 * if list is a,b,c then we have to do:
 *    v = a
 *    a = b
 *    b = c
 *    c = v
 * opposite direction is:
 *    v = c
 *    c = b
 *    b = a
 *    a = v
 */
int move[6][3] = {
  { 0, 3, 1 },
  { 1, 3, 4 },
  { 3, 6, 4 },
  { 3, 5, 6 },
  { 2, 5, 3 },
  { 0, 2, 3 },
};

typedef struct {
  int count;             /* -1 if no path exists */
  int moves[5040];       /* way too much, but I don't want to think */
                         /* stored as [1..6] for clock direction
                          * and [-1..-6] for counterclock
                          */
} moves_list_t;

moves_list_t to_solution[5040];
int packed_state_to_move_index[8*8*8*8*8*8*8];
int move_index_to_packed_state[5040];

#define printf(...) /**/ do {} while (0)

int pack_state(int state[7])
{
  return state[0] |
         (state[1] << 3) |
         (state[2] << 6) |
         (state[3] << 9) |
         (state[4] << 12) |
         (state[5] << 15) |
         (state[6] << 18);
}

void unpack_state(int state[7], int in)
{
  state[0] = in & 0x07;
  state[1] = (in >> 3) & 0x07;
  state[2] = (in >> 6) & 0x07;
  state[3] = (in >> 9) & 0x07;
  state[4] = (in >> 12) & 0x07;
  state[5] = (in >> 15) & 0x07;
  state[6] = (in >> 18) & 0x07;
}

int is_target(int state[7])
{
  return state[0] == 0
         && state[1] == 1
         && state[2] == 2
         && state[3] == 3
         && state[4] == 4
         && state[5] == 5
         && state[6] == 6;
}

void init_move_index(void)
{
  int move_index = 0;
  for (int packed_state = 0; packed_state < 2097152; packed_state++)
    packed_state_to_move_index[packed_state] = -1;
  for (int packed_state = 0; packed_state < 2097152; packed_state++) {
    int state[7];
    unpack_state(state, packed_state);
    /* all values must be in [0..6] */
    if (state[0] > 6 || state[1] > 6 || state[2] > 6 || state[3] > 6
        || state[4] > 6 || state[5] > 6 || state[6] > 6)
      continue;
    /* all must be != */
    int v = 0;
    for (int i = 0; i < 7; i++) v |= 1 << state[i];
    if (v != 127) continue;
    packed_state_to_move_index[packed_state] = move_index;
    move_index_to_packed_state[move_index] = packed_state;
    move_index++;
  }
  printf("last move_index %d\n", move_index);
}

int do_move(int ret[7], int s[7], int m, int direction)
{
  for (int i = 0; i < 7; i++) ret[i] = s[i];
  if (direction == 0) {
    int v = ret[move[m][0]];
    ret[move[m][0]] = ret[move[m][1]];
    ret[move[m][1]] = ret[move[m][2]];
    ret[move[m][2]] = v;
  } else {
    int v = ret[move[m][2]];
    ret[move[m][2]] = ret[move[m][1]];
    ret[move[m][1]] = ret[move[m][0]];
    ret[move[m][0]] = v;
  }

  /* indicate to the caller if the hole was moved */
  return ret[move[m][0]] == 3
         || ret[move[m][1]] == 3
         || ret[move[m][2]] == 3;
}

int do_move_in_place(int ret[7], int m, int direction)
{
  if (direction == 0) {
    int v = ret[move[m][0]];
    ret[move[m][0]] = ret[move[m][1]];
    ret[move[m][1]] = ret[move[m][2]];
    ret[move[m][2]] = v;
  } else {
    int v = ret[move[m][2]];
    ret[move[m][2]] = ret[move[m][1]];
    ret[move[m][1]] = ret[move[m][0]];
    ret[move[m][0]] = v;
  }

  /* indicate to the caller if the hole was moved */
  return ret[move[m][0]] == 3
         || ret[move[m][1]] == 3
         || ret[move[m][2]] == 3;
}

typedef struct {
  int previous;
  int m;
} breadth_t;

void breadth(int init_packed_state, int first_direction)
{
  int visited[5040]; for (int i = 0; i < 5040; i++) visited[i] = 0;
  breadth_t breadth[5040];
  int state[7];
  unpack_state(state, init_packed_state);

printf("digging %d %d %d %d %d %d %d\n", state[0], state[1], state[2], state[3], state[4], state[5], state[6]);

  int move_id = packed_state_to_move_index[init_packed_state];
  visited[move_id] = 1;

if (is_target(state)) printf("xx solution!\n");

  if (is_target(state)) { to_solution[move_id].count = 0; return; }

  int breadth_size = 0;

  /* put first moves */
  int next_state[7];

  for (int m = 0; m < 6; m++) {
    int hole_has_moved = do_move(next_state, state, m, first_direction);
    if (!hole_has_moved) continue;
    if (is_target(next_state)) {
      move_id = packed_state_to_move_index[init_packed_state];
      to_solution[move_id].moves[0] = (m + 1) * (first_direction == 0 ? 1 : -1);
      to_solution[move_id].count = 1;
      return;
    }
printf("  move %d leads to %d %d %d %d %d %d %d (packed %d)\n", (m+1)*(first_direction == 0 ? 1 : -1), next_state[0], next_state[1], next_state[2], next_state[3], next_state[4], next_state[5], next_state[6], pack_state(next_state));
    move_id = packed_state_to_move_index[pack_state(next_state)];
    if (visited[move_id]) continue;
    visited[move_id] = 1;
printf("visited{%d] = 1\n", move_id);
    breadth[breadth_size].m = (m + 1) * (first_direction == 0 ? 1 : -1);
printf("    breadth[%d].m = %d\n", breadth_size, breadth[breadth_size].m);
    breadth[breadth_size].previous = -1;
    breadth_size++;
  }

  /* go */
  int breadth_pos = 0;
  while (breadth_pos != breadth_size) {
    /* retrieve path */
    int path[5040];
    int path_size = 0;
    int cur = breadth_pos;
    while (cur != -1) {
      path[path_size] = breadth[cur].m;
      path_size++;
      cur = breadth[cur].previous;
    }
    /* apply path (which was retrieved in reverse) */
    unpack_state(state, init_packed_state);
printf("  start at %d %d %d %d %d %d %d\n", state[0], state[1], state[2], state[3], state[4], state[5], state[6]);
    for (int m = path_size - 1; m >= 0; m--) {
      do_move_in_place(state, abs(path[m]) - 1, path[m] < 0);
printf("  move %d leads to %d %d %d %d %d %d %d (packed %d)\n", path[m], state[0], state[1], state[2], state[3], state[4], state[5], state[6], pack_state(state));
    }

printf("  get next states from %d %d %d %d %d %d %d\n", state[0], state[1], state[2], state[3], state[4], state[5], state[6]);
    /* not reached, add next moves */
    /* alternate directions of rotation */
    int next_direction = path[0] > 0 ? 1 : 0;
    for (int m = 0; m < 6; m++) {
      int hole_has_moved = do_move(next_state, state, m, next_direction);
      if (!hole_has_moved) continue;
      /* target reached? */
      if (is_target(next_state)) {
        move_id = packed_state_to_move_index[init_packed_state];
printf("yo len %d move_id %d\n", path_size, move_id);
        to_solution[move_id].count = path_size + 1;
        /* store path, in order */
        for (int i = path_size - 1; i >= 0; i--)
          to_solution[move_id].moves[path_size - 1 - i] = path[i];
        to_solution[move_id].moves[path_size] = (m + 1) * (next_direction == 0 ? 1 : -1);
        return;
      }
printf("    move %d leads to %d %d %d %d %d %d %d (packed %d)\n", (m+1)*(next_direction == 0 ? 1 : -1), next_state[0], next_state[1], next_state[2], next_state[3], next_state[4], next_state[5], next_state[6], pack_state(next_state));
      move_id = packed_state_to_move_index[pack_state(next_state)];
      if (visited[move_id]) continue;
      visited[move_id] = 1;
printf("visited{%d] = 1\n", move_id);
      breadth[breadth_size].m = (m + 1) * (next_direction == 0 ? 1 : -1);
printf("    breadth[%d].m = %d\n", breadth_size, m);
      breadth[breadth_size].previous = breadth_pos;
      breadth_size++;
    }

    /* try next path */
    breadth_pos++;
  }

  /* solution not reached */
  move_id = packed_state_to_move_index[init_packed_state];
  to_solution[move_id].count = -1;
}

int main(void)
{
  init_move_index();

  for (int packed_state = 0; packed_state < 2097152; packed_state++) {
    if (packed_state_to_move_index[packed_state] == -1)
      continue;
    breadth(packed_state, 0);
  }

#undef printf

  printf("typedef struct {\n");
  printf("  int count;  /* -1 if no path exists */\n");
  printf("  int *moves; /* moves stored as [1..6] for clock direction\n");
  printf("               * and [-1..-6] for counterclock\n");
  printf("               */\n");
  printf("} moves_list_t;\n\n");
  printf("moves_list_t last_ring[2][5040] = {\n");
  printf("{ /* clock direction */\n");
  for (int i = 0; i < 5040; i++) {
    printf("  { .count = %d", to_solution[i].count);
    if (to_solution[i].count < 1) { printf("},\n"); continue; }
    printf(", .moves = (int[]) {");
    for (int m = 0; m < to_solution[i].count; m++) printf(" %d,", to_solution[i].moves[m]);
    printf("} },\n");
  }

  for (int packed_state = 0; packed_state < 2097152; packed_state++) {
    if (packed_state_to_move_index[packed_state] == -1)
      continue;
    breadth(packed_state, 1);
  }

  printf("},\n{ /* counterclock direction */\n");
  for (int i = 0; i < 5040; i++) {
    printf("  { .count = %d", to_solution[i].count);
    if (to_solution[i].count < 1) { printf("},\n"); continue; }
    printf(", .moves = (int[]) {");
    for (int m = 0; m < to_solution[i].count; m++) printf(" %d,", to_solution[i].moves[m]);
    printf("} },\n");
  }
  printf("} };\n");
}
