#include "solve.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

point_t position_to_pixel(int column, int line, int n)
{
  point_t ret;
  int start_x = 10 + abs(line - n + 1) * 5;

  ret.y = 10 + line * 10 * cos(M_PI / 6);
  ret.x = 10 * column + start_x;

  return ret;
}

position_t pixel_to_position(point_t pixel, int n)
{
  position_t ret;

  int line;

  ret.y = line = round((pixel.y - 10) / (10 * cos(M_PI / 6)));
  ret.x = 0;

  int start_x = 10 + abs(line - n + 1) * 5;
  ret.x = round((pixel.x - start_x) / 10);

  return ret;
}

point_t rotate(point_t center, point_t p, float angle)
{
  point_t ret;

  float cc = cos(angle);
  float ss = sin(angle);

  p.x -= center.x;
  p.y -= center.y;

  ret.x = p.x * cc - p.y * ss;
  ret.y = p.x * ss + p.y * cc;

  ret.x += center.x;
  ret.y += center.y;

  return ret;
}

static int order_columns_per_line(int line, int n)
{
  return n + n - 1 - abs(line - n + 1);
}

static int order_is_grid(int line, int column, int n)
{
  return line >= 0 && line < n * 2 - 1 && column >= 0 && column < order_columns_per_line(line, n);
}

void order(solve_stone_t *o, int n)
{
  int solution_grid[2*n-1][2*n-1];
  int used[(2*n-1)*(2*n-1)]; for (int i = 0; i < (2*n-1)*(2*n-1); i++) used[i] = 0;

  /* get solution grid */
  int stone = 1;
  for (int y = 0; y < n * 2 - 1; y++) for (int x = 0; x < n * 2 - 1; x++)
    /* skip center grid position */
    if (order_is_grid(y, x, n) && !(y == n - 1 && x == n - 1)) {
      solution_grid[y][x] = stone;
      stone++;
    }
  solution_grid[N - 1][N - 1] = 0;

  point_t center;

  center = position_to_pixel(n-1, n-1, n);

  printf("center pixel %g %g\n", center.x, center.y);

  position_t cc = pixel_to_position((point_t){ center.x, center.y }, n);
  printf("center? %d %d\n", cc.x, cc.y);

  /* order is zone by zone (zone is a triangle), far then near,
   * borders then inside
   */
  for (int zone = 0 + 3; zone < 6 + 3; zone++)
  for (int i = n-1; i >= 2; i--) {
    point_t left = (point_t){ .x = center.x + i * 10, .y = center.y };
    cc = pixel_to_position(left, n);
    printf("left? %d %d\n", cc.x, cc.y);

    point_t pa = rotate(center, left, -M_PI / 3 - zone * M_PI / 3);
    cc = pixel_to_position(pa, n);
    printf("zone %d up right? %d %d\n", zone, cc.x, cc.y);

    point_t pb = rotate(center, left, -M_PI / 3 - (zone + 1) * M_PI / 3);
    cc = pixel_to_position(pb, n);
    printf("zone %d up left? %d %d\n", zone, cc.x, cc.y);

    for (int j = 0; j <= i / 2; j++) {
      point_t pc;
      pc.x = pa.x + (pb.x - pa.x) * ((float)j / i);
      pc.y = pa.y + (pb.y - pa.y) * ((float)j / i);
      cc = pixel_to_position(pc, n);
      int stone = solution_grid[cc.y][cc.x];
      if (!used[stone]) {
        printf("put stone %d (%d %d)\n", stone, cc.x, cc.y);
        used[stone] = 1;
        o->stone = stone;
        o->x = cc.x;
        o->y = cc.y;
        o++;
      }
      point_t pd;
      pd.x = pa.x + (pb.x - pa.x) * ((float)(i - j) / i);
      pd.y = pa.y + (pb.y - pa.y) * ((float)(i - j) / i);
      cc = pixel_to_position(pd, n);
      stone = solution_grid[cc.y][cc.x];
      if (!used[stone]) {
        printf("put stone %d (%d %d)\n", stone, cc.x, cc.y);
        used[stone] = 1;
        o->stone = stone;
        o->x = cc.x;
        o->y = cc.y;
        o++;
      }
    }
  }
}

#if MAIN
int main(void)
{
  int n = 7;
  solve_stone_t o[(2*n-1)*(2*n-1)];
  order(o, n);
}
#endif
