#include "solve.h"

#include "last-ring.h"

#include <stdio.h>
#include <stdlib.h>

void get_around(position_t cur, position_t *around)
{
  /* left - right */
  around[0] = cur; around[0].x--;
  around[1] = cur; around[1].x++;

  /* up */
  around[2] = cur; around[2].y--;
  around[3] = cur; around[3].y--;
  if (cur.y < N) {
    around[2].x--;
  } else {
    around[3].x++;
  }

  /* down */
  around[4] = cur; around[4].y++;
  around[5] = cur; around[5].y++;
  if (cur.y < N - 1) {
    around[5].x++;
  } else {
    around[4].x--;
  }

  /* normalize */
  for (int i = 0; i < 6; i++) {
    point_t p = canonical_position((point_t){.x = around[i].x, .y = around[i].y });
    around[i].x = p.x;
    around[i].y = p.y;
  }
}

typedef struct {
  int previous;
  position_t position;
} breadth_search_stone_t;

typedef struct {
  int previous;
  move_t move;
} breadth_search_hole_t;

void dump(int working_grid[2*N-1][2*N-1])
{
  for (int y = 0; y < 2*N-1; y++) {
    for (int x = 0; x < 2*N-1; x++)
      if (is_grid(y, x)) printf("%d ", working_grid[y][x]);
    printf("\n");
  }
}

position_t hole_positions[2*N+1][4*N][6];
move_t hole_moves[2][2*N-1][2*N-1][6];

void generate_hole_moves_and_positions(void)
{
#if 0
memset(hole_moves, -1, sizeof(hole_moves));
#endif

  static int generated_hole_moves = 0;
  if (generated_hole_moves) return;
  generated_hole_moves = 1;

  for (int y = 0; y < N * 4 - 1; y++) {
    int columns_per_line = N + N - abs((y + 1) / 2 - N);
    for (int x = 0; x < columns_per_line; x++) {
      point_t around[3];
      positions_around(around, (point_t){.x = x, .y = y});
      /* order of moves around a position:
       *   0
       *  / \
       * 5  1
       * |  |
       * 4  2
       *  \/
       *  3
       */
      position_t a = (position_t){.x = around[0].x, .y = around[0].y};
      position_t b = (position_t){.x = around[1].x, .y = around[1].y};
      position_t c = (position_t){.x = around[2].x, .y = around[2].y};
      if (y % 2) {
        /*   \a/
         *   c|b
         */
        hole_moves[0][a.x][a.y][3] = (move_t){.direction = 0, .x = x, .y = y};
        hole_moves[0][b.x][b.y][5] = (move_t){.direction = 0, .x = x, .y = y};
        hole_moves[0][c.x][c.y][1] = (move_t){.direction = 0, .x = x, .y = y};
        hole_moves[1][a.x][a.y][3] = (move_t){.direction = 1, .x = x, .y = y};
        hole_moves[1][b.x][b.y][5] = (move_t){.direction = 1, .x = x, .y = y};
        hole_moves[1][c.x][c.y][1] = (move_t){.direction = 1, .x = x, .y = y};

        hole_positions[x][y][3] = (position_t){.x = a.x, .y = a.y};
        hole_positions[x][y][5] = (position_t){.x = b.x, .y = b.y};
        hole_positions[x][y][1] = (position_t){.x = c.x, .y = c.y};
      } else {
        /*   b|c
         *   /a\
         */
        hole_moves[0][a.x][a.y][0] = (move_t){.direction = 0, .x = x, .y = y};
        hole_moves[0][b.x][b.y][2] = (move_t){.direction = 0, .x = x, .y = y};
        hole_moves[0][c.x][c.y][4] = (move_t){.direction = 0, .x = x, .y = y};
        hole_moves[1][a.x][a.y][0] = (move_t){.direction = 1, .x = x, .y = y};
        hole_moves[1][b.x][b.y][2] = (move_t){.direction = 1, .x = x, .y = y};
        hole_moves[1][c.x][c.y][4] = (move_t){.direction = 1, .x = x, .y = y};

        hole_positions[x][y][0] = (position_t){.x = a.x, .y = a.y};
        hole_positions[x][y][2] = (position_t){.x = b.x, .y = b.y};
        hole_positions[x][y][4] = (position_t){.x = c.x, .y = c.y};
      }
    }
  }

#if 0
int count = 0;
for (int y = 0; y < 2*N-1; y++) {
  for (int x = 0; x < 2*N-1; x++) {
    for (int d = 0; d < 2; d++) {
      for (int m = 0; m < 6; m++) {
        if (hole_moves[d][x][y][m].direction >= 0)
          printf("hole_moves[%d][%d][%d][%d] = direction %d x %d y %d\n", d, x, y, m, hole_moves[d][x][y][m].direction, hole_moves[d][x][y][m].x, hole_moves[d][x][y][m].y), count++;
      }
    }
  }
  printf("\n");
}
printf("count %d count/2/6 %d\n", count, count/2/6);
#endif
}

void get_hole_moves(move_t *moves, int direction, int x, int y)
{
  for (int i = 0; i < 6; i++)
    moves[i] = hole_moves[direction][x][y][i];
}

position_t hole_position_from_move(move_t m, int hole_id)
{
  return hole_positions[m.x][m.y][hole_id];
}

#define set_grid_and_position(p, v) do { \
    working_grid[(int)p.y][(int)p.x] = v; \
    stones_position[v].x = p.x; \
    stones_position[v].y = p.y; \
  } while (0)
#define get_working_grid(p) working_grid[(int)p.y][(int)p.x]
#define DO_MOVE(m) do { \
    point_t p[3]; \
    positions_around(p, (point_t){.x = (m).x, .y = (m).y}); \
    if ((m).direction == 0) { \
      /* left */ \
      int v = get_working_grid(p[0]); \
      set_grid_and_position(p[0], get_working_grid(p[2])); \
      set_grid_and_position(p[2], get_working_grid(p[1])); \
      set_grid_and_position(p[1], v); \
    } else { \
      /* right */ \
      int v = get_working_grid(p[0]); \
      set_grid_and_position(p[0], get_working_grid(p[1])); \
      set_grid_and_position(p[1], get_working_grid(p[2])); \
      set_grid_and_position(p[2], v); \
    } \
  } while (0)
#define UNDO_MOVE(m) do { \
    point_t p[3]; \
    positions_around(p, (point_t){.x = (m).x, .y = (m).y}); \
    if ((m).direction == 1) { \
      /* undo right is do left */ \
      int v = get_working_grid(p[0]); \
      set_grid_and_position(p[0], get_working_grid(p[2])); \
      set_grid_and_position(p[2], get_working_grid(p[1])); \
      set_grid_and_position(p[1], v); \
    } else { \
      /* undo left is do right */ \
      int v = get_working_grid(p[0]); \
      set_grid_and_position(p[0], get_working_grid(p[1])); \
      set_grid_and_position(p[1], get_working_grid(p[2])); \
      set_grid_and_position(p[2], v); \
    } \
  } while (0)

#define push(m) do { \
printf("push %d %d %d\n", m.direction, m.x, m.y); \
    ret.size++; \
    move_t *new_moves = realloc(ret.moves, ret.size * sizeof(move_t)); \
    if (new_moves == NULL) { free(ret.moves); return (moves_t){.moves= (void *)0, .size = 0}; } \
    ret.moves = new_moves; \
    ret.moves[ret.size - 1].direction = m.direction; \
    ret.moves[ret.size - 1].x = m.x; \
    ret.moves[ret.size - 1].y = m.y; \
  } while (0)

/*****************************************************************/
/* basic solve                                                   */
/*****************************************************************/

moves_t basic_solve(int start_direction,
    int stone_count,
    position_t stones_position[(2 * N - 1) * (2 * N - 1)],
    int working_grid[2*N-1][2*N-1],
    position_t stones_solution_position[(2 * N - 1) * (2 * N - 1)],
    int solution_grid[2 * N - 1][2 * N - 1],
    solve_stone_t ordered_list[stone_count - 6])
{
  moves_t ret = (moves_t){.moves = (void *)0, .size = 0};

  int is_fixed[stone_count]; for (int i = 0; i < stone_count; i++) is_fixed[i] = 0;

  int i;
  for (i = 0; i < stone_count - 6; i++) {
    printf("putting stone %d at its place\n", ordered_list[i].stone);
    int already_visited[2*N-1][2*N-1]; for (int i = 0; i < N * 2 - 1; i++) for (int j = 0; j < N * 2 - 1; j++) already_visited[i][j] = 0;
    int stone = ordered_list[i].stone;
    position_t start, end, cur;
    /* find a shortest path for the stone */
    breadth_search_stone_t breadth[(2*N-1)*(2*N-1)];
    int breadth_cur;
    int breadth_size = 0;
    start = stones_position[stone];
    end =  stones_solution_position[stone];
    printf("move stone %d from %d.%d to %d.%d\n", stone, start.y, start.x, end.y, end.x);
    if (start.x == end.x && start.y == end.y) { is_fixed[stone - 1] = 1; continue; }
    cur = start;
    position_t around[6];
    already_visited[cur.y][cur.x] = 1;
    get_around(cur, around);
    for (int i = 0; i < 6; i++) printf("around %d.%d is %d.%d\n", cur.y, cur.x, around[i].y, around[i].x);
    for (int i = 0; i < 6; i++) {
      int around_stone = working_grid[around[i].y][around[i].x];
      if (already_visited[around[i].y][around[i].x]
          || (around_stone != 0 && is_fixed[around_stone - 1]))
        continue;
      already_visited[around[i].y][around[i].x] = 1;
      breadth[breadth_size].previous = -1;
      breadth[breadth_size].position = (position_t){.x = around[i].x, .y = around[i].y};
      breadth_size++;
    }
    /* breadth first search */
    position_t stone_path[(2*N-1)*(2*N-1)];
    int stone_path_size = 0;
    breadth_cur = 0;
    while (1) {
      if (breadth_cur == breadth_size) break;
      cur = breadth[breadth_cur].position;
      if (!(cur.x == end.x && cur.y == end.y)) {
        get_around(cur, around);
        for (int i = 0; i < 6; i++) {
          int around_stone = working_grid[around[i].y][around[i].x];
          if (already_visited[around[i].y][around[i].x]
              || (around_stone != 0 && is_fixed[around_stone - 1]))
            continue;
          already_visited[around[i].y][around[i].x] = 1;
          breadth[breadth_size].previous = breadth_cur;
          breadth[breadth_size].position = (position_t){.x = around[i].x, .y = around[i].y};
          breadth_size++;
        }
        breadth_cur++;
        continue;
      }
      /* cur == end, let's extract the path */
      while (1) {
        if (breadth_cur == -1) break;
        stone_path[stone_path_size] = breadth[breadth_cur].position;
        stone_path_size++;
        breadth_cur = breadth[breadth_cur].previous;
      }
      for (int i = stone_path_size - 1; i >= 0; i--) {
        printf("stone %d pass through %d.%d\n", stone, stone_path[i].y, stone_path[i].x);
      }
      break;
    }
    /* do we have a path? if no, stop solving here and return what we get so far */
    if (stone_path_size == 0) {
      printf("no path found for stone %d\n", stone);
      break;
    }
    /* we have the path, try to run it */
    breadth_search_hole_t hole_breadth[12 * (N*2-1) * (N*2-1)];
    int hole_breadth_size;
    cur = start;
    position_t next;
    for (int i = stone_path_size - 1; i >= 0; i--) {
      next = stone_path[i];
      printf("move stone %d to %d.%d\n", stone, next.x, next.y);
      /* breadth-first move the hole until 'stone' is at position 'next' */
      hole_breadth_size = 0;
      int hole_visited[2][6][2*N-1][2*N-1];
      for (int i = 0; i < 2; i++) for (int j = 0; j < 6; j++) for (int k = 0; k < 2*N-1; k++) for (int l = 0; l < 2*N-1; l++) hole_visited[i][j][k][l] = 0;
      /* first move's direction can be what we want, let's try both */
      move_t moves[12];
      int init_moves_count = 6;
      /* if some moves already computed, start_direction is fixed */
      if (ret.size)
        start_direction = 1 - ret.moves[ret.size - 1].direction;
      get_hole_moves(moves, start_direction == -1 ? 0 : start_direction, stones_position[0].x, stones_position[0].y);
      if (start_direction == -1) {
        get_hole_moves(moves+6, 1, stones_position[0].x, stones_position[0].y);
        init_moves_count = 12;
      }
      for (int m = 0; m < init_moves_count; m++) {
        /* check validity (all stones must be movable) */
        point_t moving_stones[3];
        positions_around(moving_stones, (point_t){.x = moves[m].x, .y = moves[m].y}); 
        int valid_move = 1;
        for (int i = 0; i < 3; i++) {
          int move_stone = working_grid[(int)moving_stones[i].y][(int)moving_stones[i].x];
          if (move_stone != 0 && is_fixed[move_stone - 1]) {
             valid_move = 0;
             break;
          }
          /* the stone is only allowed to move to 'next' */
          if (move_stone == stone) {
            DO_MOVE(moves[m]);
            if (!(stones_position[stone].x == next.x && stones_position[stone].y == next.y)) {
              valid_move = 0;
              UNDO_MOVE(moves[m]);
              break;
            }
            UNDO_MOVE(moves[m]);
          }
        }
        if (valid_move) {
          hole_breadth[hole_breadth_size].previous = -1;
          hole_breadth[hole_breadth_size].move = moves[m];
          /* to understand hole id, see comment in generate_hole_moves_and_positions() which returns moves in their id's order */
          int hole_id = m % 6;
          position_t hole_position = hole_position_from_move(moves[m], hole_id);
          hole_visited[moves[m].direction][hole_id][hole_position.x][hole_position.y] = 1;
          hole_breadth_size++;
        }
      }
      /* breadth search */
      int current_hole_move = 0;
      while (1) {
        if (current_hole_move == hole_breadth_size) break;
        /* get current path from starting hole position to current one */
        move_t current_path[12 * (N*2-1) * (N*2-1)];
        int current_path_size = 0;
        int current_hole = current_hole_move;
        while (current_hole != -1) {
          current_path[current_path_size] = hole_breadth[current_hole].move;
          current_path_size++;
          current_hole = hole_breadth[current_hole].previous;
        }
        /* do moves of this path */
        for (int i = current_path_size - 1; i >= 0; i--) {
          DO_MOVE(current_path[i]);
        }
        /* stop if stone is at its position */
        if (stones_position[stone].x == next.x && stones_position[stone].y == next.y) {
          printf("stone %d is at %d %d\n", stone, next.x, next.y);
          /* save path */
          for (int i = current_path_size - 1; i >= 0; i--)
            push(current_path[i]);
//printf("ret %d\n", ret.size);
//for (int i = 0; i < ret.size; i++) printf(" [ %d %d %d ]", ret.moves[i].direction, ret.moves[i].x, ret.moves[i].y);
//printf("\n");
//return ret;
          break;
        }
        /* get following moves */
        /* path is in reverse, so last move of path is at current_path[0] */
        get_hole_moves(moves, 1 - current_path[0].direction, stones_position[0].x, stones_position[0].y);
        for (int m = 0; m < 6; m++) {
          /* skip move if we already visited it */
          int hole_id = m;
          position_t hole_position = hole_position_from_move(moves[m], hole_id);
          if (hole_visited[moves[m].direction][hole_id][hole_position.x][hole_position.y] == 1)
            continue;
          /* check validity (all stones must be movable) */
          point_t moving_stones[3];
          positions_around(moving_stones, (point_t){.x = moves[m].x, .y = moves[m].y}); 
          int valid_move = 1;
          for (int i = 0; i < 3; i++) {
            int move_stone = working_grid[(int)moving_stones[i].y][(int)moving_stones[i].x];
            if (move_stone != 0 && is_fixed[move_stone - 1]) {
               valid_move = 0;
               break;
            }
            /* the stone is only allowed to move to 'next' */
            if (move_stone == stone) {
              DO_MOVE(moves[m]);
              if (!(stones_position[stone].x == next.x && stones_position[stone].y == next.y)) {
                valid_move = 0;
                UNDO_MOVE(moves[m]);
                break;
              }
              UNDO_MOVE(moves[m]);
            }
          }
          if (valid_move) {
            hole_breadth[hole_breadth_size].previous = current_hole_move;
            hole_breadth[hole_breadth_size].move = moves[m];
            int hole_id = m;
            position_t hole_position = hole_position_from_move(moves[m], hole_id);
            hole_visited[moves[m].direction][hole_id][hole_position.x][hole_position.y] = 1;
            hole_breadth_size++;
          }
        }
        /* undo moves of this path */
        for (int i = 0; i < current_path_size; i++) {
          UNDO_MOVE(current_path[i]);
        }
        current_hole_move++;
      }
      /* one move done for the stone, do next */
      cur = next;
    }
    /* stone should be at its target position - if not, finish */
    if (!(stones_position[stone].x == next.x && stones_position[stone].y == next.y)) break;
    is_fixed[stone - 1] = 1;
  }

  ret.next_stone = i == stone_count - 6 ? -1 : i;
printf("solve returns %d moves (next_stone %d)\n", ret.size, ret.next_stone);

  return ret;
}

/* return -1 on error, 0 if ok */
int append_moves(moves_t *to, moves_t *from)
{
  move_t *m = realloc(to->moves, (to->size + from->size) * sizeof(move_t));
  if (m == (void *)0) { printf("error: out of memory\n"); return -1; }
  to->moves = m;
  for (int i = 0; i < from->size; i++) to->moves[to->size + i] = from->moves[i];
  to->size += from->size;
  return 0;
}

void dump_moves(moves_t *m)
{
  printf("moves[%d]:", m->size);
  for (int i = 0; i < m->size; i++) printf(" [%d %d %d]", m->moves[i].direction, m->moves[i].x, m->moves[i].y);
  printf("\n");
}

moves_t solve(int start_direction)
{
  generate_hole_moves_and_positions();

  position_t stones_position[(2 * N - 1) * (2 * N - 1)];
  int working_grid[2*N-1][2*N-1];
  position_t stones_solution_position[(2 * N - 1) * (2 * N - 1)];
  int solution_grid[2 * N - 1][2 * N - 1];

  for (int i = 0; i < 2*N-1; i++) for (int j = 0; j < 2*N-1; j++)
    working_grid[i][j] = grid[i][j];

  int stone_count = 0;
  for (int y = 0; y < N * 2 - 1; y++) for (int x = 0; x < N * 2 - 1; x++)
    if (is_grid(y, x)) stone_count++;
  stone_count--;
  printf("%d stones\n", stone_count);

  /* get stones and hole positions */
  for (int y = 0; y < N * 2 - 1; y++) for (int x = 0; x < N * 2 - 1; x++)
    if (is_grid(y, x)) {
      stones_position[grid[y][x]].x = x;
      stones_position[grid[y][x]].y = y;
    }

  printf("hole at %d %d\n", stones_position[0].y, stones_position[0].x);

  /* get solution grid */
  int stone = 1;
  for (int y = 0; y < N * 2 - 1; y++) for (int x = 0; x < N * 2 - 1; x++)
    /* skip center grid position */
    if (is_grid(y, x) && !(y == N - 1 && x == N - 1)) {
      solution_grid[y][x] = stone;
      stones_solution_position[stone].x = x;
      stones_solution_position[stone].y = y;
      stone++;
    }
  solution_grid[N - 1][N - 1] = 0;

  for (int y = 0; y < N * 2 - 1; y++) {
    for (int x = 0; x < N * 2 - 1; x++) {
      if (is_grid(y, x)) printf("%d ", solution_grid[y][x]);
    }
    printf("\n");
  }

#if 1
  /* get ordered list of stones - will be solved in that order */
  /* solve outer ring, then inner ring, etc. last ring (6 stones) not solved */
  solve_stone_t ordered_list[stone_count - 6];
  int cur_stone = 0;
  for (int ring = 0;; ring++) {
    for (int y = ring; y < N * 2 - 1 - ring; y++)
    for (int x = ring; x < columns_per_line(y) - ring; x++)
      if (is_grid(y, x)
          && ((y == ring || y == N * 2 - 1 - 1 - ring)
               || (x == ring || x == columns_per_line(y) - 1 - ring))) {
        ordered_list[cur_stone].stone = solution_grid[y][x];
        ordered_list[cur_stone].x = x;
        ordered_list[cur_stone].y = y;
        cur_stone++;
        if (cur_stone == stone_count - 6) {
          /* reverse solving order of two last stones - it's easier */
          solve_stone_t v = ordered_list[cur_stone - 2];
          ordered_list[cur_stone - 2] = ordered_list[cur_stone - 1];
          ordered_list[cur_stone - 1] = v;
          goto got_ordered_list;
        }
      }
    /* reverse solving order of two last stones - it's easier */
    solve_stone_t v = ordered_list[cur_stone - 2];
    ordered_list[cur_stone - 2] = ordered_list[cur_stone - 1];
    ordered_list[cur_stone - 1] = v;
  }
got_ordered_list:
#else
  solve_stone_t ordered_list[stone_count - 6];
  void order(solve_stone_t *o, int n);
  order(ordered_list, N);
#endif

  for (int i = 0; i < stone_count - 6; i++)
    printf("we have to put stone %d at %d.%d\n",
           ordered_list[i].stone, ordered_list[i].y, ordered_list[i].x);

  int next_stone = 0;
  moves_t ret = { .moves = (void *)0, .size = 0 };

printf("before first basic_solve\n");
dump(working_grid);
  moves_t basic = basic_solve(start_direction,
                              stone_count,
                              stones_position,
                              working_grid,
                              stones_solution_position,
                              solution_grid,
                              ordered_list);
#if 0
  if (basic.size == 0) {
printf("first basic returns size = 0\n");
    return basic;
  }
#endif

printf("after first basic_solve\n");
dump(working_grid);
printf("basic gives moves: ");
dump_moves(&basic);

printf("ret moves before append: ");
dump_moves(&ret);
  if (append_moves(&ret, &basic) == -1) return ret;
printf("ret moves after append: ");
dump_moves(&ret);

  /* stop if all stones done */
  if (basic.next_stone == -1 && basic.size)
    return ret;

  next_stone = basic.next_stone;
if (next_stone == -1) next_stone = 0;

keep_going:
printf("xxx basic.next_stone %d\n", basic.next_stone);

  /* get all possible (useful or not) paths of length 1..3 after basic solver */

  breadth_search_hole_t hole_breadth[12 * (N*2-1) * (N*2-1)]; /* way too big */
  int hole_breadth_size;

  move_t moves[6];
  int direction;
  if (ret.size == 0) direction = start_direction;
  else direction = 1 - ret.moves[ret.size - 1].direction;
  get_hole_moves(moves, direction, stones_position[0].x, stones_position[0].y);
  for (int i = 0; i < 6; i++) {
    hole_breadth[i].move = moves[i];
    hole_breadth[i].previous = -1;
  }
  hole_breadth_size = 6;

  /* breadth (limited to max path length 3) */
  /* we try all possible moves */
  int current_hole_move = 0;
  while (current_hole_move < hole_breadth_size) {
    move_t current_path[3];
    int current_path_size = 0;
    int current_hole = current_hole_move;
    while (current_hole != -1) {
      current_path[current_path_size] = hole_breadth[current_hole].move;
      current_path_size++;
      current_hole = hole_breadth[current_hole].previous;
    }
    /* do moves of this path */
    for (int i = current_path_size - 1; i >= 0; i--) {
      DO_MOVE(current_path[i]);
    }
    /* add next move if path length is < 3 */
    if (current_path_size < 3) {
      move_t moves[6];
      /* last move of path is at 0 */
      int direction = 1 - current_path[0].direction;
      get_hole_moves(moves, direction, stones_position[0].x, stones_position[0].y);
      for (int i = 0; i < 6; i++) {
        hole_breadth[hole_breadth_size + i].move = moves[i];
        hole_breadth[hole_breadth_size + i].previous = current_hole_move;
      }
      hole_breadth_size += 6;
    }
    /* undo moves of this path */
    for (int i = 0; i < current_path_size; i++) {
      UNDO_MOVE(current_path[i]);
    }
    current_hole_move++;
  }

  /* try all path until we can move farther than 'next_stone' */
  for (current_hole_move = 0; current_hole_move < hole_breadth_size; current_hole_move++) {
    move_t current_path[3];
    int current_path_size = 0;
    int current_hole = current_hole_move;
    while (current_hole != -1) {
      current_path[current_path_size] = hole_breadth[current_hole].move;
      current_path_size++;
      current_hole = hole_breadth[current_hole].previous;
    }
    /* do moves of this path */
printf("before next path of len %d (start hole %d)\n", current_path_size, current_hole_move);
dump(working_grid);
    for (int i = current_path_size - 1; i >= 0; i--) {
printf("doing move %d %d %d\n", current_path[i].direction, current_path[i].x, current_path[i].y);
      DO_MOVE(current_path[i]);
    }
    /* try the basic solver, see if it moves */
    int saved_working_grid[2*N-1][2*N-1];
    position_t saved_stones_position[(2 * N - 1) * (2 * N - 1)];
    for (int i = 0; i < 2*N-1; i++) for (int j = 0; j < 2*N-1; j++)
      saved_working_grid[i][j] = working_grid[i][j];
    for (int i = 0; i < (N * 2 - 1) * (N * 2 - 1); i++)
      saved_stones_position[i] = stones_position[i];
printf("before next basic_solve\n");
dump(working_grid);
    free(basic.moves);
    int direction = 1 - current_path[0].direction;
    basic = basic_solve(direction,
                        stone_count,
                        stones_position,
                        working_grid,
                        stones_solution_position,
                        solution_grid,
                        ordered_list);
printf("after next basic_solve (basic.next_stone %d next_stone %d\n", basic.next_stone, next_stone);
dump(working_grid);
printf("basic gives moves: ");
dump_moves(&basic);
    //if ((basic.next_stone == -1 && basic.size > 0) || basic.next_stone > next_stone) {
    if (basic.size > 0 && (basic.next_stone == -1 || basic.next_stone > next_stone)) {
printf("ret moves before append: ");
dump_moves(&ret);
      moves_t moves_of_path;
      move_t the_moves[current_path_size];
      for (int i = 0; i < current_path_size; i++) the_moves[i] = current_path[current_path_size - 1 - i];
      moves_of_path.moves = the_moves;
      moves_of_path.size = current_path_size;
      if (append_moves(&ret, &moves_of_path) == -1) return ret;
      if (append_moves(&ret, &basic) == -1) return ret;
printf("ret moves after append: ");
dump_moves(&ret);
printf("moved, basic.next_stone %d next_stone %d (ret.size %d)\n", basic.next_stone, next_stone, ret.size);
      next_stone = basic.next_stone;
      /* stop if all done */
      if (basic.next_stone == -1) return ret;
      goto keep_going;
    }
    free(basic.moves);
    basic.moves = (void *)0;
    for (int i = 0; i < 2*N-1; i++) for (int j = 0; j < 2*N-1; j++)
      working_grid[i][j] = saved_working_grid[i][j];
    for (int i = 0; i < (N * 2 - 1) * (N * 2 - 1); i++)
      stones_position[i] = saved_stones_position[i];
    /* undo moves of this path */
    for (int i = 0; i < current_path_size; i++) {
      UNDO_MOVE(current_path[i]);
    }
printf("after next path of len %d (hole %d) (ret.size %d)\n", current_path_size, current_hole_move, ret.size);
dump(working_grid);
  }

  /* we couldn't move on, we're done */
  return ret;
}
