#include "solve2.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "stack.h"

typedef struct {
  int grid[2*N-1][2*N-1];
  position_t stone_position[(2*N-1)*(2*N-1)];
} pos_t;

typedef struct {
  move_t move;
  int hole_id;
} move_and_hole_id_t;

typedef struct {
  int previous;
  move_and_hole_id_t move;
} breadth_hm_t;

static int position_equal(position_t a,position_t b){return a.x==b.x&&a.y==b.y;}

static void append_moves(moves_t *to, moves_t *from)
{
  if (to->size + from->size == 0) return;
  move_t *m = realloc(to->moves, (to->size + from->size) * sizeof(move_t));
  if (m == (void *)0) abort();
  to->moves = m;
  for (int i = 0; i < from->size; i++) to->moves[to->size + i] = from->moves[i];
  to->size += from->size;
}

#define set_grid_and_position(pos, v) do { \
    p->grid[(int)pos.y][(int)pos.x] = v; \
    p->stone_position[v].x = pos.x; \
    p->stone_position[v].y = pos.y; \
  } while (0)
#define get_working_grid(pos) p->grid[(int)pos.y][(int)pos.x]
#define DO_MOVE(m) do { \
    point_t pt[3]; \
    positions_around(pt, (point_t){.x = (m).x, .y = (m).y}); \
    if ((m).direction == 0) { \
      /* left */ \
      int v = get_working_grid(pt[0]); \
      set_grid_and_position(pt[0], get_working_grid(pt[2])); \
      set_grid_and_position(pt[2], get_working_grid(pt[1])); \
      set_grid_and_position(pt[1], v); \
    } else { \
      /* right */ \
      int v = get_working_grid(pt[0]); \
      set_grid_and_position(pt[0], get_working_grid(pt[1])); \
      set_grid_and_position(pt[1], get_working_grid(pt[2])); \
      set_grid_and_position(pt[2], v); \
    } \
  } while (0)
#define UNDO_MOVE(m) do { \
    point_t pt[3]; \
    positions_around(pt, (point_t){.x = (m).x, .y = (m).y}); \
    if ((m).direction == 1) { \
      /* undo right is do left */ \
      int v = get_working_grid(pt[0]); \
      set_grid_and_position(pt[0], get_working_grid(pt[2])); \
      set_grid_and_position(pt[2], get_working_grid(pt[1])); \
      set_grid_and_position(pt[1], v); \
    } else { \
      /* undo left is do right */ \
      int v = get_working_grid(pt[0]); \
      set_grid_and_position(pt[0], get_working_grid(pt[1])); \
      set_grid_and_position(pt[1], get_working_grid(pt[2])); \
      set_grid_and_position(pt[2], v); \
    } \
  } while (0)

int get_hole_moves2(pos_t *p, move_and_hole_id_t *out, int *hole_busy,
                    int position_busy[2*N-1][2*N-1], int direction,
                    int *stone_busy, int stone)
{
  void get_hole_moves(move_t *moves, int direction, int x, int y);
  move_t pm[6];
  get_hole_moves(pm, direction, p->stone_position[0].x, p->stone_position[0].y);

  int ret = 0;

  /* check validity (all stones must be movable) */
  for (int m = 0; m < 6; m++) {
    /* move must not be busy */
    int hole_id = m;
    position_t hole_position_from_move(move_t m, int hole_id);
    position_t hole_position = hole_position_from_move(pm[m], hole_id);
    if (hole_busy[direction*6*(2*N-1)*(2*N-1) +
                  hole_id*(2*N-1)*(2*N-1) +
                  hole_position.x*(2*N-1) +
                  hole_position.y])
      continue;
    point_t moving_stones[3];
    /* if 'stone' moves, it shall not go somewhere it already went */
    position_t old_pos = p->stone_position[stone];
    DO_MOVE(pm[m]);
    position_t new_pos = p->stone_position[stone];
    UNDO_MOVE(pm[m]);
    if (!position_equal(old_pos, new_pos) && position_busy[new_pos.x][new_pos.y])
      continue;
    /* all stones must be movable */
    positions_around(moving_stones, (point_t){.x = pm[m].x, .y = pm[m].y}); 
    int i;
    for (i = 0; i < 3; i++) {
      int move_stone = p->grid[(int)moving_stones[i].y][(int)moving_stones[i].x];
      if (move_stone != 0 && stone_busy[move_stone]) break;
    }
    if (i == 3) {
      /* all is movable, move is good */
      out[ret].move = pm[m];
      out[ret].hole_id = m;
      ret++;
    }
  }

  return ret;
}

/* return path size
 * beware: out_path is in reverse!
 */
int get_path(move_and_hole_id_t *out_path, breadth_hm_t *bhm, int start)
{
  int ret = 0;
  while (start != -1) {
    out_path[ret] = bhm[start].move;
    start = bhm[start].previous;
    ret++;
  }
  return ret;
}

void do_path(pos_t *p, move_and_hole_id_t *path, int path_size, moves_t *ret,
             int *hole_busy, int stone, int position_busy[2*N-1][2*N-1])
{
  position_t stone_pos = p->stone_position[stone];

  position_busy[p->stone_position[stone].x][p->stone_position[stone].y] = 1;
  for (int i = path_size - 1; i >= 0; i--) {
    DO_MOVE(path[i].move);
    int hole_id = path[i].hole_id;
    position_t hole_position_from_move(move_t m, int hole_id);
    position_t hole_position = hole_position_from_move(path[i].move, hole_id);
    int direction = path[i].move.direction;
    hole_busy[direction*6*(2*N-1)*(2*N-1) +
              hole_id*(2*N-1)*(2*N-1) +
              hole_position.x*(2*N-1) +
              hole_position.y] = 1;
    position_busy[p->stone_position[stone].x][p->stone_position[stone].y] = 1;
    if (!position_equal(stone_pos, p->stone_position[stone])) {
      memset(hole_busy, 0, sizeof(int) * 2*6 * (2*N-1)*(2*N-1));
      stone_pos = p->stone_position[stone];
    }
  }
}

void undo_path(pos_t *p, move_and_hole_id_t *path, int path_size, moves_t *ret,
               int *hole_busy, int stone)
{
  memset(hole_busy, 0, sizeof(int) * 2*6 * (2*N-1)*(2*N-1));
  for (int i = 0; i < path_size; i++)
    UNDO_MOVE(path[i].move);
}

  breadth_hm_t bhm[2 * 6 * (2*N-1) * (2*N-1) * (2*N-1)*(2*N-1)];
moves_t solve_stone(int stone, pos_t *cur, pos_t *next, pos_t *solution,
                    int direction, int *stone_busy)
{
  moves_t ret = { .moves = NULL, .size = 0 };
  pos_t p = *cur;

  if (position_equal(p.stone_position[stone], solution->stone_position[stone]))
    return ret;

  int hole_busy[2*6 * (2*N-1)*(2*N-1)]; memset(hole_busy, 0, sizeof(hole_busy));
  int position_busy[2*N-1][2*N-1]; memset(position_busy, 0, sizeof(position_busy));

  /* get initial moves */
  move_and_hole_id_t hm[6];
  int hm_size = get_hole_moves2(&p, hm, hole_busy, position_busy,
                                direction, stone_busy, stone);


  /* store initial moves in the breadth search list */
  int bhm_size = 0;

  for (int i = 0; i < hm_size; i++) {
    bhm[bhm_size].previous = -1;
    bhm[bhm_size].move = hm[i];
    bhm_size++;
  }

  /* breadth search */
  for (int i = 0; i < bhm_size; i++) {
    move_and_hole_id_t path[(2*N-1)*(2*N-1)*2];
    /* do the path */
    int path_size = get_path(path, bhm, i);
printf("testing path size %d\n", path_size);
    do_path(&p, path, path_size, &ret, hole_busy, stone, position_busy);
    if (position_equal(p.stone_position[stone], solution->stone_position[stone])) {
      *cur = p;
      ret.size = path_size;
      ret.moves = malloc(path_size * sizeof(move_t)); if (ret.moves == NULL) abort();
      for (int m = path_size - 1; m >= 0; m--)
        ret.moves[m] = path[path_size - 1 - m].move;
      return ret;
    }
    /* get next moves */
    direction = 1 - path[0].move.direction;
    hm_size = get_hole_moves2(&p, hm, hole_busy, position_busy,
                              direction, stone_busy, stone);
    for (int j = 0; j < hm_size; j++) {
printf("bhm_size %d\n", bhm_size);
      bhm[bhm_size].previous = i;
      bhm[bhm_size].move = hm[j];
      bhm_size++;
    }
    /* undo the path */
    undo_path(&p, path, path_size, &ret, hole_busy, stone);
  }

  return ret;
}

moves_t solve2(int start_direction)
{
  pos_t pos_vars[3]; memset(pos_vars, 0, sizeof(pos_vars));
  pos_t *current_position = &pos_vars[0];
  pos_t *next_position = &pos_vars[1];
  pos_t solution_position;

  {
    void generate_hole_moves_and_positions(void);
    static int generated = 0;
    if (!generated) {
      generated = 1;
      generate_hole_moves_and_positions();
    }
  }

  /* get stone count */
  int stone_count = 0;
  for (int y = 0; y < N * 2 - 1; y++) for (int x = 0; x < N * 2 - 1; x++)
    if (is_grid(y, x)) stone_count++;
  stone_count--;

  /* get stone order */
  solve_stone_t stone_order[stone_count - 6];
  void order(solve_stone_t *o, int n);
  order(stone_order, N);

  /* get solution grid */
  int stone = 1;
  for (int y = 0; y < N * 2 - 1; y++) for (int x = 0; x < N * 2 - 1; x++)
    /* skip center grid position */
    if (is_grid(y, x) && !(y == N - 1 && x == N - 1)) {
      solution_position.grid[y][x] = stone;
      solution_position.stone_position[stone].x = x;
      solution_position.stone_position[stone].y = y;
      stone++;
    }
  solution_position.grid[N - 1][N - 1] = 0;

  /* get initial position */
  memcpy(current_position->grid, grid, sizeof(grid));
  /* get stones and hole positions */
  for (int y = 0; y < N * 2 - 1; y++) for (int x = 0; x < N * 2 - 1; x++)
    if (is_grid(y, x)) {
      current_position->stone_position[grid[y][x]].x = x;
      current_position->stone_position[grid[y][x]].y = y;
    }

  moves_t ret = { .moves = NULL, .size = 0 };
  int direction = start_direction;

  int stone_busy[stone_count+1]; memset(stone_busy, 0, sizeof(stone_busy));

  for (int stone_index = 0; stone_index < stone_count - 6; stone_index++) {
    int stone = stone_order[stone_index].stone;
//if (stone == 12) return ret;
    printf("solving stone %d (go to %d %d)\n", stone, solution_position.stone_position[stone].x, solution_position.stone_position[stone].y);
    moves_t m = solve_stone(stone, current_position, next_position,
                            &solution_position, direction, stone_busy);
    append_moves(&ret, &m);
    free(m.moves);
printf("after append_moves ret.size %d m.size %d\n", ret.size, m.size);
    /* stop if stone is not at its position */
    if (!position_equal(current_position->stone_position[stone],
                        solution_position.stone_position[stone])) {
      printf("stone %d could not be solved\n", stone);
      ret.next_stone = stone;
      return ret;
    }
    printf("stone %d solved\n", stone);
    stone_busy[stone] = 1;
    if (ret.size)
      direction = 1 - ret.moves[ret.size - 1].direction;
  }

  printf("all stones solved\n");
  ret.next_stone = -1;
  return ret;
}
