#include "solve3.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define printf(...) do {} while(0)

typedef struct {
  int grid[2*N-1][2*N-1];
  position_t stone_position[(2*N-1)*(2*N-1)];
} pos_t;

typedef struct {
  move_t move;
  int hole_id;
} move_and_hole_id_t;

typedef struct {
  int previous;
  move_and_hole_id_t move;
} breadth_hm_t;

typedef struct {
  position_t from;
  position_t to;
  int direction;
  int last_direction;
  int path_len;
  int previous;
  int index;
} bfs_item_t;

typedef struct {
  bfs_item_t *l;
  int *lorder;
  int size;
  int maxsize;
  int next;
} bfs_t;

void dump(int working_grid[2*N-1][2*N-1]);

void bfs_init(bfs_t *b)
{
printf("bfs_init()\n");
  b->l = NULL;
  b->lorder = NULL;
  b->maxsize = b->size = 0;
  b->next = 0;
}

void bfs_free(bfs_t *b)
{
  free(b->l);
  free(b->lorder);
}

int bfs_over(bfs_t *b)
{
  return b->next >= b->size;
}

bfs_item_t bfs_get(bfs_t *b)
{
  int index = b->lorder[b->next];
  bfs_item_t ret = b->l[index];
  b->next++;
  return ret;
}

void bfs_dump(bfs_t *b)
{
  printf("bfs[%d].l = ", b->size);
  for (int i = 0; i < b->size; i++) printf(" [%d %d (%d)]", b->l[i].index, b->l[i].previous, b->l[i].path_len);
  printf("\n");
  printf("bfs[%d].lorder = ", b->size);
  for (int i = 0; i < b->size; i++) printf(" [%d]", b->lorder[i]);
  printf("\n");
}

void bfs_add(bfs_t *b, position_t from, position_t to, int direction,
             int last_direction, int path_len, int previous)
{
  if (b->size == b->maxsize) {
    b->maxsize += 4096;
    b->l = realloc(b->l, b->maxsize * sizeof(bfs_item_t)); if (b->l == NULL) abort();
    b->lorder = realloc(b->lorder, b->maxsize * sizeof(int)); if (b->lorder == NULL) abort();
  }
  int pos = b->size;
  b->l[pos] = (bfs_item_t){.from = from, .to = to, .path_len = path_len,
                           .direction = direction,
                           .last_direction = last_direction,
                           .previous = previous,
                           .index = pos};
  b->lorder[pos] = pos;
  /* keep order of list: increasing path length */
  while (pos && b->l[b->lorder[pos]].path_len < b->l[b->lorder[pos-1]].path_len) {
    /* swap */
    int v = b->lorder[pos];
    b->lorder[pos] = b->lorder[pos-1];
    b->lorder[pos-1] = v;
    pos--;
  }
  b->size++;
}

static int position_equal(position_t a,position_t b){return a.x==b.x&&a.y==b.y;}

static void append_moves(moves_t *to, moves_t *from)
{
  if (from->size == 0) return;
  move_t *m = realloc(to->moves, (to->size + from->size) * sizeof(move_t));
  if (m == (void *)0) abort();
  to->moves = m;
  for (int i = 0; i < from->size; i++) to->moves[to->size + i] = from->moves[i];
if (from->size && to->size) if (to->moves[to->size - 1].direction == from->moves[0].direction) abort();
  to->size += from->size;
}

typedef struct {
  int previous;
  move_t move;
} breadth_search_hole_t;

#define set_grid_and_position(p, pos, v) do { \
    (p)->grid[(int)pos.y][(int)pos.x] = v; \
    (p)->stone_position[v].x = pos.x; \
    (p)->stone_position[v].y = pos.y; \
  } while (0)
#define get_working_grid(p, pos) (p)->grid[(int)pos.y][(int)pos.x]
#define DO_MOVE(p, m) do { \
    point_t pt[3]; \
    positions_around(pt, (point_t){.x = (m).x, .y = (m).y}); \
    if ((m).direction == 0) { \
      /* left */ \
      int v = get_working_grid(p, pt[0]); \
      set_grid_and_position(p, pt[0], get_working_grid(p, pt[2])); \
      set_grid_and_position(p, pt[2], get_working_grid(p, pt[1])); \
      set_grid_and_position(p, pt[1], v); \
    } else { \
      /* right */ \
      int v = get_working_grid(p, pt[0]); \
      set_grid_and_position(p, pt[0], get_working_grid(p, pt[1])); \
      set_grid_and_position(p, pt[1], get_working_grid(p, pt[2])); \
      set_grid_and_position(p, pt[2], v); \
    } \
  } while (0)
#define UNDO_MOVE(p, m) do { \
    point_t pt[3]; \
    positions_around(pt, (point_t){.x = (m).x, .y = (m).y}); \
    if ((m).direction == 1) { \
      /* undo right is do left */ \
      int v = get_working_grid(p, pt[0]); \
      set_grid_and_position(p, pt[0], get_working_grid(p, pt[2])); \
      set_grid_and_position(p, pt[2], get_working_grid(p, pt[1])); \
      set_grid_and_position(p, pt[1], v); \
    } else { \
      /* undo left is do right */ \
      int v = get_working_grid(p, pt[0]); \
      set_grid_and_position(p, pt[0], get_working_grid(p, pt[1])); \
      set_grid_and_position(p, pt[1], get_working_grid(p, pt[2])); \
      set_grid_and_position(p, pt[2], v); \
    } \
  } while (0)

position_t hole_position_from_move(move_t m, int hole_id);

static moves_t get_hole_path(int stone, pos_t *cur, int direction,
                             int *stone_busy, int last_direction,
                             position_t target)
{
  pos_t p = *cur;
  breadth_search_hole_t hole_breadth[12 * (N*2-1) * (N*2-1)];
  int hole_breadth_size = 0;
  int hole_visited[2][6][2*N-1][2*N-1]; memset(hole_visited, 0, sizeof(hole_visited));
  move_t moves[6];
  void get_hole_moves(move_t *moves, int direction, int x, int y);
  get_hole_moves(moves, direction, p.stone_position[0].x, p.stone_position[0].y);
  for (int m = 0; m < 6; m++) {
    /* check validity (all stones must be movable) */
    point_t moving_stones[3];
    positions_around(moving_stones, (point_t){.x = moves[m].x, .y = moves[m].y}); 
    int valid_move = 1;
    for (int i = 0; i < 3; i++) {
      int move_stone = p.grid[(int)moving_stones[i].y][(int)moving_stones[i].x];
      if (move_stone != 0 && stone_busy[move_stone]) {
         valid_move = 0;
         break;
      }
#if 1
      /* the stone is only allowed to move to 'next' */
      if (move_stone == stone) {
        DO_MOVE(&p, moves[m]);
        if (!position_equal(p.stone_position[stone], target)) { //(stones_position[stone].x == next.x && stones_position[stone].y == next.y)) {
          valid_move = 0;
          UNDO_MOVE(&p, moves[m]);
          break;
        }
        UNDO_MOVE(&p, moves[m]);
      }
#endif
    }
    if (valid_move) {
      hole_breadth[hole_breadth_size].previous = -1;
      hole_breadth[hole_breadth_size].move = moves[m];
      /* to understand hole id, see comment in generate_hole_moves_and_positions() which returns moves in their id's order */
      int hole_id = m % 6;
      position_t hole_position = hole_position_from_move(moves[m], hole_id);
      hole_visited[moves[m].direction][hole_id][hole_position.x][hole_position.y] = 1;
      hole_breadth_size++;
    }
  }
  int current_hole_move = 0;
  int current_path_size;
  move_t current_path[12 * (N*2-1) * (N*2-1)];
  while (1) {
    current_path_size = 0;
    if (current_hole_move == hole_breadth_size) break;
    /* get current path from starting hole position to current one */
    int current_hole = current_hole_move;
    while (current_hole != -1) {
      current_path[current_path_size] = hole_breadth[current_hole].move;
      current_path_size++;
      current_hole = hole_breadth[current_hole].previous;
    }
    /* do moves of this path */
    for (int i = current_path_size - 1; i >= 0; i--) {
      DO_MOVE(&p, current_path[i]);
    }
    /* stop if stone is at its position and last direction is good */
    if (position_equal(p.stone_position[stone], target)
        && last_direction == current_path[0].direction) {
#if 0
      printf("stone %d is at %d %d\n", stone, next.x, next.y);
      /* save path */
      for (int i = current_path_size - 1; i >= 0; i--)
        push(current_path[i]);
#endif
      break;
    }
    /* get following moves */
    /* path is in reverse, so last move of path is at current_path[0] */
    get_hole_moves(moves, 1 - current_path[0].direction, p.stone_position[0].x, p.stone_position[0].y);
    for (int m = 0; m < 6; m++) {
      /* skip move if we already visited it */
      int hole_id = m;
      position_t hole_position = hole_position_from_move(moves[m], hole_id);
      if (hole_visited[moves[m].direction][hole_id][hole_position.x][hole_position.y] == 1)
        continue;
      /* check validity (all stones must be movable) */
      point_t moving_stones[3];
      positions_around(moving_stones, (point_t){.x = moves[m].x, .y = moves[m].y}); 
      int valid_move = 1;
      for (int i = 0; i < 3; i++) {
        int move_stone = p.grid[(int)moving_stones[i].y][(int)moving_stones[i].x];
        if (move_stone != 0 && stone_busy[move_stone]) {
           valid_move = 0;
           break;
        }
#if 1
        /* the stone is only allowed to move to 'next' */
        if (move_stone == stone) {
          DO_MOVE(&p, moves[m]);
          if (!position_equal(p.stone_position[stone], target)) {
            valid_move = 0;
            UNDO_MOVE(&p, moves[m]);
            break;
          }
          UNDO_MOVE(&p, moves[m]);
        }
#endif
      }
      if (valid_move) {
        hole_breadth[hole_breadth_size].previous = current_hole_move;
        hole_breadth[hole_breadth_size].move = moves[m];
        int hole_id = m;
        position_t hole_position = hole_position_from_move(moves[m], hole_id);
        hole_visited[moves[m].direction][hole_id][hole_position.x][hole_position.y] = 1;
        hole_breadth_size++;
      }
    }
    /* undo moves of this path */
    for (int i = 0; i < current_path_size; i++) {
      UNDO_MOVE(&p, current_path[i]);
    }
    current_hole_move++;
  }

  moves_t ret;
  ret.moves = NULL;
  ret.size = current_path_size;
  if (ret.size == 0) return ret;
  ret.moves = malloc(sizeof(move_t) * ret.size); if (ret.moves == NULL) abort();
  for (int i = ret.size - 1; i >= 0; i--)
    ret.moves[ret.size - 1 - i] = current_path[i];
printf("hole move size %d\n", current_path_size);
  return ret;
}

static int get_length(int stone, pos_t *cur, int direction, int *stone_busy,
                      int last_direction, position_t target)
{
  moves_t m = get_hole_path(stone, cur, direction, stone_busy,
                            last_direction, target);
  int ret = m.size;
  free(m.moves);
printf("current_path_size %d\n", ret);
  return ret;
}

void append_move(moves_t *ret, move_t m)
{
  int inc = 4096;
  if ((ret->size & (inc - 1)) == 0) {
    ret->moves = realloc(ret->moves, (ret->size + inc) * sizeof(move_t)); if (ret->moves == NULL) abort();
  }
  ret->moves[ret->size] = m;
  ret->size++;
}

static void do_path(pos_t *pos, bfs_t *b, bfs_item_t p, int *stone_busy, moves_t *ret)
{
printf("start do_path\n");
  bfs_item_t *l = NULL;
  int size = 0;
  int maxsize = 0;
  int index = p.index;
  while (index != -1) {
    if (size == maxsize) {
      maxsize += 1024;
      l = realloc(l, maxsize * sizeof(bfs_item_t)); if (l == NULL) abort();
if (maxsize > 1000000) { printf("uhoh\n"); exit(1); }
    }
    l[size] = b->l[index];
    index = b->l[index].previous;
    size++;
  }
printf("do path[%d] =", size);
for (int i = size - 1; i >= 0; i--) printf(" [(%d %d) (%d %d) %d]", l[i].from.y, l[i].from.x, l[i].to.y, l[i].to.x, l[i].last_direction);
printf("\n");
  for (int i = size - 1; i >= 0; i--) {
    int stone = pos->grid[l[i].from.y][l[i].from.x];
    moves_t m = get_hole_path(stone, pos, l[i].direction, stone_busy,
                              l[i].last_direction, l[i].to);
printf("i %d hole moves len %d\n", i, m.size);
    for (int j = 0; j < m.size; j++) {
      DO_MOVE(pos, m.moves[j]);
      append_move(ret, m.moves[j]);
    }
    free(m.moves);
  }
  free(l);

#if 0
  if (p.previous != -1) do_path(pos, b, b->l[p.previous], stone_busy, ret);
  int stone = pos->grid[p.from.y][p.from.x];
  moves_t m = get_hole_path(stone, pos, p.direction, stone_busy,
                            p.last_direction, p.to);
  for (int i = 0; i < m.size; i++) {
    DO_MOVE(pos, m.moves[i]);
    append_move(ret, m.moves[i]);
  }
  free(m.moves);
#endif
}

static void undo_path(pos_t *pos, moves_t *m)
{
  for (int i = m->size - 1; i >= 0; i--)
    UNDO_MOVE(pos, m->moves[i]);
}

static void dump_busy(int *v)
{
  for (int i = 0; i < 37; i++) printf("%d", v[i]);
  printf("\n");
}

static  int transition_busy[2*N-1][2*N-1][2*N-1][2*N-1][2];  /* x1,y1  x2,y2  direction */
static moves_t solve_stone(int stone, pos_t *cur, pos_t *solution,
                           int direction, int *stone_busy)
{
//printf("solve_stone stone %d direction %d\n", stone, direction);
//fprintf(stdout, "solve_stone stone %d direction %d\n", stone, direction);
  moves_t ret = (moves_t){.moves = NULL, .size = 0};
  if (position_equal(cur->stone_position[stone], solution->stone_position[stone]))
    return (moves_t){.moves = NULL, .size = 0};

//printf("init pos:\n"); dump(cur->grid);

  memset(transition_busy, 0, sizeof(transition_busy));

  /* get first transitions */
  void get_around(position_t cur, position_t *around);
  position_t around[6];
  position_t stone_position = cur->stone_position[stone];
  get_around(stone_position, around);

  bfs_t b;
  bfs_init(&b);
//bfs_dump(&b);

  for (int i = 0; i < 6; i++) {
    int len = get_length(stone, cur, direction, stone_busy, 0, around[i]);
    if (len > 0) bfs_add(&b, stone_position, around[i], direction, 0, len, -1);
    len = get_length(stone, cur, direction, stone_busy, 1, around[i]);
    if (len > 0) bfs_add(&b, stone_position, around[i], direction, 1, len, -1);
  }

printf("start the while in solve_stone\n");
  while (1) {
printf("in while b.size %d\n", b.size);
//printf("at start of while(1):\n"); dump(cur->grid);
    if (bfs_over(&b)) break;
    bfs_item_t p = bfs_get(&b);
    do_path(cur, &b, p, stone_busy, &ret);
    if (position_equal(cur->stone_position[stone], solution->stone_position[stone]))
      break;
//printf("after do_path (ret.size %d):\n", ret.size); dump(cur->grid);

if (ret.size != p.path_len) exit(1);
    {
      int x1 = p.from.x;
      int y1 = p.from.y;
      int x2 = p.to.x;
      int y2 = p.to.y;
      int direction = p.last_direction;
      int path_len = p.path_len;
      if (transition_busy[x1][y1][x2][y2][direction] > 0
          && transition_busy[x1][y1][x2][y2][direction] < path_len)
        goto dont_process;
    }

    stone_position = cur->stone_position[stone];
    get_around(stone_position, around);

if (ret.size == 0) abort();
    direction = 1 - ret.moves[ret.size - 1].direction;

    for (int i = 0; i < 6; i++) {
if (around[i].y == 6 && around[i].x == 2)
  printf("yooooooo\n");
      int x1 = stone_position.x;
      int y1 = stone_position.y;
      int x2 = around[i].x;
      int y2 = around[i].y;
      int previous_len = p.path_len;
      //if (!transition_busy[x1][y1][x2][y2][0]) {
      {
printf("adding %d %d to %d %d\n", x1, y1, x2, y2);
//printf("before adding\n"); bfs_dump(&b);
//dump(cur->grid);
//dump_busy(stone_busy);
        int len = get_length(stone, cur, direction, stone_busy, 0, around[i]);
printf("len %d\n", len);
        if (len > 0
            && (transition_busy[x1][y1][x2][y2][0] == 0
                || len + previous_len < transition_busy[x1][y1][x2][y2][0])) {
          bfs_add(&b, stone_position, around[i], direction, 0, len + previous_len, p.index);
          transition_busy[x1][y1][x2][y2][0] = len + previous_len;
        }
      }
      //if (!transition_busy[x1][y1][x2][y2][1]) {
      {
printf("adding %d %d to %d %d\n", x1, y1, x2, y2);
//printf("before adding\n"); bfs_dump(&b);
        int len = get_length(stone, cur, direction, stone_busy, 1, around[i]);
printf("len %d\n", len);
        if (len > 0
            && (transition_busy[x1][y1][x2][y2][1] == 0
                || len + previous_len < transition_busy[x1][y1][x2][y2][1])) {
          bfs_add(&b, stone_position, around[i], direction, 1, len + previous_len, p.index);
          transition_busy[x1][y1][x2][y2][1] = len + previous_len;
        }
      }
    }

dont_process:
    undo_path(cur, &ret);
//printf("after undo_path:\n"); dump(cur->grid);
    free(ret.moves);
    ret.moves = NULL;
    ret.size = 0;
  }

  bfs_free(&b);

  return ret;
}

  pos_t pos_vars[1]; //memset(pos_vars, 0, sizeof(pos_vars));
  pos_t *current_position = &pos_vars[0];
  pos_t solution_position;
moves_t solve3(int start_direction)
{
  memset(pos_vars, 0, sizeof(pos_vars));

  {
    void generate_hole_moves_and_positions(void);
    static int generated = 0;
    if (!generated) {
      generated = 1;
      generate_hole_moves_and_positions();
    }
  }

  /* get stone count */
  int stone_count = 0;
  for (int y = 0; y < N * 2 - 1; y++) for (int x = 0; x < N * 2 - 1; x++)
    if (is_grid(y, x)) stone_count++;
  stone_count--;

  /* get solution grid */
  int stone = 1;
  for (int y = 0; y < N * 2 - 1; y++) for (int x = 0; x < N * 2 - 1; x++)
    /* skip center grid position */
    if (is_grid(y, x) && !(y == N - 1 && x == N - 1)) {
      solution_position.grid[y][x] = stone;
      solution_position.stone_position[stone].x = x;
      solution_position.stone_position[stone].y = y;
      stone++;
    }
  solution_position.grid[N - 1][N - 1] = 0;

#if 0
/* alternative solution (abandoned, did not prove better) */
int sol[] = {
          0, 1, 3, 4, 5, 6, 
        7, 2, 9, 10, 11, 12, 13, 
      14, 8, 15, 16, 18, 19, 20, 21, 
    22, 23, 24, 17, 26, 27, 28, 29, 30, 
  31, 32, 33, 25, 34, 35, 37, 38, 39, 40, 
41, 42, 43, 44, 45, 36, 46, 47, 48, 49, 50, 
  51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 
    61, 62, 63, 64, 65, 66, 67, 68, 69, 
      70, 71, 72, 73, 74, 75, 76, 77, 
        78, 79, 80, 81, 82, 83, 84, 
          85, 86, 87, 88, 89, 90, 
};
int sol_pos = 0;
  for (int y = 0; y < N * 2 - 1; y++) for (int x = 0; x < N * 2 - 1; x++)
    if (is_grid(y, x)) {
      int stone = sol[sol_pos];
      solution_position.grid[y][x] = stone;
      solution_position.stone_position[stone].x = x;
      solution_position.stone_position[stone].y = y;
      sol_pos++;
    }
#endif
dump(solution_position.grid);

#if 1
  /* get ordered list of stones - will be solved in that order */
  /* solve outer ring, then inner ring, etc. last ring (6 stones) not solved */
  solve_stone_t stone_order[stone_count - 6];
  int cur_stone = 0;
  for (int ring = 0;; ring++) {
    for (int y = ring; y < N * 2 - 1 - ring; y++)
    for (int x = ring; x < columns_per_line(y) - ring; x++)
      if (is_grid(y, x)
          && ((y == ring || y == N * 2 - 1 - 1 - ring)
               || (x == ring || x == columns_per_line(y) - 1 - ring))) {
        stone_order[cur_stone].stone = solution_position.grid[y][x];
        stone_order[cur_stone].x = x;
        stone_order[cur_stone].y = y;
        cur_stone++;
        if (cur_stone == stone_count - 6) {
          /* reverse solving order of two last stones - it's easier */
          solve_stone_t v = stone_order[cur_stone - 2];
          stone_order[cur_stone - 2] = stone_order[cur_stone - 1];
          stone_order[cur_stone - 1] = v;
          goto got_stone_order;
        }
      }
    /* reverse solving order of two last stones - it's easier */
    solve_stone_t v = stone_order[cur_stone - 2];
    stone_order[cur_stone - 2] = stone_order[cur_stone - 1];
    stone_order[cur_stone - 1] = v;
  }
got_stone_order:
#else
  /* get stone order */
  solve_stone_t stone_order[stone_count - 6];
  void order(solve_stone_t *o, int n);
  order(stone_order, N);
#endif

#if 0
/* alternative order, abandoned, did not prove better) */
point_t position_to_pixel(int column, int line, int n);
position_t pixel_to_position(point_t pixel, int n);
point_t p0 = position_to_pixel(0, 0, 6);
fprintf(stdout, "p0 %g %g\n", p0.x, p0.y);
cur_stone = 0;
for (int ring = 10; ring >= 0; ring--) {
  point_t p = { .x = p0.x + ring * 5, .y = p0.y + ring * 10 * cos(M_PI/6) };
  position_t pos = pixel_to_position(p, 6);
  stone_order[cur_stone].stone = solution_position.grid[pos.y][pos.x];
  stone_order[cur_stone].x = pos.x;
  stone_order[cur_stone].y = pos.y;
  cur_stone++;
fprintf(stdout, "center %g %g\n", p.x, p.y);
  for (int i = 0; i < 5; i++) {
    point_t pixel;
    pixel.x = p.x - (5 - i) * 10;
    pixel.y = p.y;
    position_t pos = pixel_to_position(pixel, 6);
fprintf(stdout, "in line: %d: pos x/y %d %d (pixel %g %g)\n", i, pos.x, pos.y, pixel.x, pixel.y);
    if (is_grid(pos.y, pos.x)) {
      stone_order[cur_stone].stone = solution_position.grid[pos.y][pos.x];
      stone_order[cur_stone].x = pos.x;
      stone_order[cur_stone].y = pos.y;
      cur_stone++;
        if (cur_stone == stone_count - 6) goto ddd;
    }
  }
  for (int i = 6; i >= 1; i--) {
    point_t pixel;
    pixel.x = p.x + i * 5;
    pixel.y = p.y - i * 10 * cos(M_PI / 6);
    position_t pos = pixel_to_position(pixel, 6);
    if (is_grid(pos.y, pos.x)) {
      stone_order[cur_stone].stone = solution_position.grid[pos.y][pos.x];
      stone_order[cur_stone].x = pos.x;
      stone_order[cur_stone].y = pos.y;
      cur_stone++;
        if (cur_stone == stone_count - 6) goto ddd;
    }
  }
}
ddd:
fprintf(stdout, "at end cur_stone %d\n", cur_stone); fflush(stdout);
for (int i = 0; i < cur_stone; i++) fprintf(stdout, " %d", stone_order[i].stone);
printf(stdout, "\n");
#endif

  /* get initial position */
  memcpy(current_position->grid, grid, sizeof(grid));
  /* get stones and hole positions */
  for (int y = 0; y < N * 2 - 1; y++) for (int x = 0; x < N * 2 - 1; x++)
    if (is_grid(y, x)) {
      current_position->stone_position[grid[y][x]].x = x;
      current_position->stone_position[grid[y][x]].y = y;
    }

  moves_t ret = { .moves = NULL, .size = 0 };
  int direction = start_direction;

  int stone_busy[stone_count+1]; memset(stone_busy, 0, sizeof(stone_busy));

  for (int stone_index = 0; stone_index < stone_count - 6; stone_index++) {
    int stone = stone_order[stone_index].stone;
//if (stone == 12) return ret;
    printf("[%d/%d] solving stone %d (go to %d %d)\n", stone_index, stone_count - 6, stone, solution_position.stone_position[stone].x, solution_position.stone_position[stone].y);
    fprintf(stdout, "[%d/%d] solving stone %d (go to %d %d)\n", stone_index, stone_count - 6, stone, solution_position.stone_position[stone].x, solution_position.stone_position[stone].y);
    moves_t m = solve_stone(stone, current_position,
                            &solution_position, direction, stone_busy);
printf("solve_stone(%d) returns %d moves\n", stone, m.size);
void dump_moves(moves_t *);
//dump_moves(&m);
//dump_moves(&ret);
    append_moves(&ret, &m);
if (ret.size) printf("last direction for stone %d = %d\n", stone, ret.moves[ret.size - 1].direction);
    free(m.moves);
    /* stop if stone is not at its position */
    if (!position_equal(current_position->stone_position[stone],
                        solution_position.stone_position[stone])) {
      printf("stone %d could not be solved\n", stone);
      ret.next_stone = stone;
      return ret;
    }
    printf("stone %d solved\n", stone);
    stone_busy[stone] = 1;
    if (m.size)
      direction = 1 - ret.moves[ret.size - 1].direction;
  }

  printf("all stones solved\n");
  ret.next_stone = -1;

  return ret;
}
