#include <stdio.h>
#include <stdlib.h>

#include "solve.h"

unsigned int grid[2*N-1][2*N-1];

int columns_per_line(int line)
{
  return N + N - 1 - abs(line - N + 1);
}

int is_grid(int line, int column)
{
  return line >= 0 && line < N * 2 - 1 && column >= 0 && column < columns_per_line(line);
}

void init_grid(void)
{
  int init[] = { INIT };
  int p = 0;
  for (int line = -1; line < 2 * N; line++)
    for (int column = -1; column < 2 * N; column++) {
      if (!is_grid(line, column)) continue;
      grid[line][column] = init[p];
      p++;
    }
}

point_t canonical_position(point_t p)
{
  if (is_grid(p.y, p.x)) return p;

  /* upper horizontal line */
  if (p.y == -1 && p.x >= 0)
    return (point_t) { .x = p.x, .y = (N - 1) * 2 };

  /* upper left line */
  if (p.x == -1 && p.y < N - 1)
    return (point_t) { .x = columns_per_line(p.y + N) - 1, .y = p.y + N };

  /* lower left line */
  if (p.x == -1 && p.y <= (N - 1) * 2)
    return (point_t) { .x = columns_per_line(p.y - (N - 1)) - 1, .y = p.y - (N - 1) };

  /* lower horizontal line */
  if (p.y == (N - 1) * 2 + 1 && p.x < N - 1)
    return (point_t) { .x = p.x + 1, .y = 0 };

  /* upper right line */
  if (p.y <= N - 1)
    return (point_t) { .x = 0, .y = p.y + N - 1 };

  /* lower right line */
  return (point_t) { .x = 0, .y = p.y - N };
}

/*
 * return point around empty cell
 *
 *    5
 *    /\
 *  4/  \0
 *   |  |
 *   |  |
 *  3\  /1
 *    \/
 *    2
 */
int get_positions_around(point_t *around, point_t p)
{
  int ret = -1;
  int ax, ay;
  int bx, by;
  int cx, cy;
  int a, b, c;

  if ((int)p.y % 2) {
    /* point to bottom
     * \a/
     * c|b
     */
    /* get coordinates (x, y) of a */
    ay = ((int)p.y - 1) / 2 - 1;
    ax = p.x;
    if (ay < N - 1)
      ax--;
    point_t p = canonical_position((point_t) { ax, ay });
    around[0] = p;
    a = grid[(int)p.y][(int)p.x];
    /* get b */
    by = ay + 1;
    bx = ax;
    if (ay < N - 1)
      bx++;
    p = canonical_position((point_t) { bx, by });
    around[1] = p;
    b = grid[(int)p.y][(int)p.x];
    /* get c */
    cx = bx - 1;
    cy = by;
    p = canonical_position((point_t) { cx, cy });
    around[2] = p;
    c = grid[(int)p.y][(int)p.x];
    if (a == 0) ret = 2;
    if (b == 0) ret = 4;
    if (c == 0) ret = 0;
  } else {
    /* points to top
     * b|c
     * /a\
     */
    /* get a */
    ay = (int)p.y / 2;
    ax = p.x;
    if (ay > N - 1)
      ax--;
    point_t p = canonical_position((point_t) { ax, ay });
    around[0] = p;
    a = grid[(int)p.y][(int)p.x];
    /* get b */
    by = ay - 1;
    bx = ax;
    if (ay < N)
      bx--;
    p = canonical_position((point_t) { bx, by });
    around[1] = p;
    b = grid[(int)p.y][(int)p.x];
    /* get c */
    cx = bx + 1;
    cy = by;
    p = canonical_position((point_t) { cx, cy });
    around[2] = p;
    c = grid[(int)p.y][(int)p.x];
    if (a == 0) ret = 5;
    if (b == 0) ret = 1;
    if (c == 0) ret = 3;
  }
fprintf(stderr, "a %d (x/y %d %d) b %d (x/y %d %d) c %d (x/y %d %d)\n", a, ax, ay, b, bx, by, c, cx, cy);
  return ret;
}

void set_grid(point_t p, int v)
{
  grid[(int)p.y][(int)p.x] = v;
}

int get_grid(point_t p)
{
  return grid[(int)p.y][(int)p.x];
}

int main(void)
{
  init_grid();

  printf("%d : ", N);
  while (1) {
    int rotation, x, y;
    if (scanf("%d %d %d\n", &rotation, &x, &y) != 3) break;
    fprintf(stderr, "%d %d %d\n", rotation, x, y);
    point_t p[3];
    int move = get_positions_around(p, (point_t) { x, y });
    if (rotation == 0) {
      int v = get_grid(p[0]);
      set_grid(p[0], get_grid(p[2]));
      set_grid(p[2], get_grid(p[1]));
      set_grid(p[1], v);
      printf("%d", move + 1);
    } else {
      int v = get_grid(p[0]);
      set_grid(p[0], get_grid(p[1]));
      set_grid(p[1], get_grid(p[2]));
      set_grid(p[2], v);
      printf("%c", move + 'A');
    }
  }
  printf("\n");

for (int y = 0; y < 2 * N - 1; fprintf(stderr, "\n"), y++)
for (int x = 0; x < 2 * N - 1; x++)
if (is_grid(y, x)) fprintf(stderr, " %d", grid[y][x]);
  return 0;
}
